#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
import threading
from abc import ABC, ABCMeta, abstractmethod
from collections import defaultdict
from threading import Condition, Event, RLock
from time import time
from typing import Any, Dict, List, Set, TypeVar

from ..pdi.asc2_req import Asc2Req
from ..pdi.base3_component import RouteComponent
from ..pdi.constants import PdiCommand
from ..pdi.irda_req import IrdaReq
from ..pdi.pdi_req import PdiReq
from ..pdi.stm2_req import Stm2Req
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import (
    BROADCAST_ADDRESS,
    CommandScope,
    CommandSyntax,
)
from ..protocol.tmcc1.tmcc1_constants import (
    TMCC1HaltCommandEnum,
)
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum as Switch
from ..utils.text_utils import title
from .comp_data import CompData, CompDataMixin

log = logging.getLogger(__name__)

C = TypeVar("C", bound=CommandDefEnum)
E = TypeVar("E", bound=CommandDefEnum)
P = TypeVar("P", bound=PdiReq)
L = TypeVar("L", bound=CommandReq)

BIG_NUMBER = float("inf")


# noinspection PyUnresolvedReferences
class ComponentState(ABC, CompDataMixin):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, scope: CommandScope = None) -> None:
        from .component_state_store import DependencyCache

        super().__init__()
        self._cv: Condition = Condition(RLock())
        self._ev = Event()
        self._is_known: bool = False
        self._scope = scope
        self._last_command: CommandReq | None = None
        self._last_command_bytes = None
        self._last_updated: float | None = None
        self._road_name = None
        self._road_number = None
        self._number = None
        self._address: int | None = None
        self._spare_1: int | None = None
        self._dependencies = DependencyCache.build()

    def __repr__(self) -> str:
        if self.is_comp_data_record is True and not self.payload:
            return str(self.comp_data)
        nm = f" {self.road_name}" if self.road_name else ""
        nu = f" #{self.road_number}" if self.road_number else ""
        pk = f" {self.payload}" if self.payload else ""
        if self.scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
            return f"{self.scope.title} {self._address:04}:{pk}{nm}{nu}"
        else:
            return f"{self.scope.title} {self._address:>2}:{pk}{nm}{nu}"

    def __lt__(self, other):
        return self.address < other.address

    def results_in(self, command: CommandReq) -> Set[E]:
        effects = self._dependencies.results_in(command.command, dereference_aliases=True, include_aliases=False)
        if command.is_data:
            # noinspection PyTypeChecker
            effects.update(
                self._dependencies.results_in(
                    (command.command, command.data), dereference_aliases=True, include_aliases=False
                )
            )
        return effects

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @property
    def friendly_scope(self) -> str:
        return self.scope.name.title()

    @property
    def address(self) -> int:
        return self._address

    @property
    def tmcc_id(self) -> int:
        return self.address

    @property
    def last_command(self) -> CommandReq:
        return self._last_command

    @property
    def last_updated(self) -> float:
        return self._last_updated

    @property
    def synchronizer(self) -> Condition:
        return self._cv

    @property
    def changed(self) -> threading.Event:
        return self._ev

    @property
    def road_name(self) -> str | None:
        return self._road_name

    @property
    def road_number(self) -> str | None:
        return self._road_number

    @property
    def name(self) -> str:
        the_name = " #" + self.road_number if self.road_number else ""
        the_name = self.road_name + the_name if self.road_name else "NA"
        return the_name

    @property
    def payload(self) -> str:
        return ""

    @property
    def spare_1(self) -> int:
        return self._spare_1

    @abstractmethod
    def update(self, command: L | P) -> None:
        from ..pdi.base_req import BaseReq
        from ..pdi.block_req import BlockReq
        from ..pdi.d4_req import D4Req

        self.changed.clear()
        if command and hasattr(command, "command") and command.command == TMCC1HaltCommandEnum.HALT:
            pass
        else:
            if self._address is None and command.address != BROADCAST_ADDRESS:
                self._address = command.address
            # invalid states
            elif self._address is None and command.address == BROADCAST_ADDRESS:
                raise AttributeError(
                    f"Received broadcast address for {self.friendly_scope} but component has not "
                    f"been initialized {self}"
                )
            elif command.address not in {self._address, BROADCAST_ADDRESS}:
                raise AttributeError(
                    f"{self.friendly_scope} #{self._address} received update for "
                    f"{command.scope.name.title()} #{command.address}, ignoring"
                )
            if self.scope != command.scope:
                scope = command.scope.name.title()
                raise AttributeError(f"{self.friendly_scope} {self.address} received update for {scope}, ignoring")
            if (
                (isinstance(command, BaseReq) and command.status == 0)
                or isinstance(command, IrdaReq)
                or isinstance(command, BlockReq)
                or isinstance(command, D4Req)
            ):
                if hasattr(command, "name") and command.name:
                    self._road_name = title(command.name)
                if hasattr(command, "number") and command.number:
                    self._road_number = command.number
                    # support lookup by road number
                    if self.road_number:
                        try:
                            from .component_state_store import ComponentStateStore

                            rn = int(self.road_number)
                            if (
                                rn > 99 >= self.address >= 1
                                and ComponentStateStore.get_state(self.scope, rn, False) is None
                            ):
                                ComponentStateStore.set_state(self.scope, rn, self)
                        except ValueError:
                            pass
            if isinstance(command, PdiReq):
                self._is_known = True
                if hasattr(command, "spare_1"):
                    self._spare_1 = command.spare_1
        self._last_updated = time()
        self._last_command = command

    @property
    def last_updated_ago(self) -> float:
        if self._last_updated is not None:
            return time() - self._last_updated
        else:
            return BIG_NUMBER

    @staticmethod
    def time_delta(last_updated: float, recv_time: float) -> float:
        if last_updated is None or recv_time is None:
            return BIG_NUMBER
        return last_updated - recv_time

    @staticmethod
    def update_aux_state(
        aux: CommandDefEnum,
        on: CommandDefEnum,
        opt1: CommandDefEnum,
        off: CommandDefEnum,
    ):
        return on if aux is None or aux in {opt1, off} else off

    @property
    def syntax(self) -> CommandSyntax:
        return CommandSyntax.LEGACY if self.is_legacy is True else CommandSyntax.TMCC

    @property
    def is_known(self) -> bool:
        """
        Returns True if the component's state is known, False otherwise.
        """
        return self._is_known

    def as_bytes(self) -> bytes | list[bytes]:
        from ..pdi.base_req import BaseReq

        """
        Returns the component state as a bytes object representative of the TMCC/Legacy
        byte sequence used to trigger the corresponding action(s) when received by the
        component.

        Used to synchronizer component state when client connects to the server.
        """
        with self.synchronizer:
            byte_str = BaseReq(self.address, PdiCommand.BASE_MEMORY, scope=self.scope, state=self).as_bytes
            return byte_str

    def _update_comp_data(self, comp_data: CompData):
        self._comp_data = comp_data
        self._comp_data_record = True

    def _harvest_effect(self, effects: Set[E]) -> E | tuple[E, int] | None:
        for effect in effects:
            if isinstance(effect, tuple):
                effect_enum = effect[0]
                effect_data = effect[1]
            else:
                effect_enum = effect
                effect_data = None
            if effect_enum.syntax == self.syntax:
                if effect_data is None:
                    return effect_enum
                else:
                    return effect_enum, effect_data
        return None

    @staticmethod
    def _as_label(prop: Any) -> str:
        return f"{prop if prop is not None else 'NA'}"

    def _as_dict(self) -> Dict[str, Any]:
        return {
            "tmcc_id": self.tmcc_id,
            "road_name": self.road_name,
            "road_number": self.road_number,
            "scope": self.scope.name.lower(),
        }

    @property
    @abstractmethod
    def is_tmcc(self) -> bool:
        """
        Returns True if the component responds to TMCC protocol, False otherwise.
        """
        ...

    @property
    @abstractmethod
    def is_legacy(self) -> bool:
        """
        Returns True if the component responds to Legacy/TMCC2 protocol, False otherwise.
        """
        ...

    @property
    @abstractmethod
    def is_lcs(self) -> bool:
        """
        Returns True if the component is an LCS device, False otherwise.
        """
        ...

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        """
        Returns the component state as a dict object containing the current state
        of the component
        """
        ...


class TmccState(ComponentState, ABC):
    __metaclass__ = ABCMeta

    def __init__(self, scope: CommandScope = None) -> None:
        super().__init__(scope)

    @property
    def is_tmcc(self) -> bool:
        return True

    @property
    def is_legacy(self) -> bool:
        return False

    @property
    def is_lcs(self) -> bool:
        return False


class LcsState(ComponentState, ABC):
    __metaclass__ = ABCMeta

    def __init__(self, scope: CommandScope = None) -> None:
        super().__init__(scope)

    @property
    def is_tmcc(self) -> bool:
        return False

    @property
    def is_legacy(self) -> bool:
        return False

    @property
    def is_lcs(self) -> bool:
        return True


class SwitchState(TmccState):
    """
    Maintain the perceived state of a Switch
    """

    def __init__(self, scope: CommandScope = CommandScope.SWITCH) -> None:
        if scope != CommandScope.SWITCH:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._state: Switch | None = None
        self._routes: set[RouteState] = set()

    def update(self, command: L | P) -> None:
        if command:
            if command.command == TMCC1HaltCommandEnum.HALT:
                return
            with self.synchronizer:
                super().update(command)
                if isinstance(command, CompDataMixin) and command.is_comp_data_record:
                    self._update_comp_data(command.comp_data)
                elif isinstance(command, CommandReq):
                    if command.command != Switch.SET_ADDRESS:
                        self._state = command.command
                elif isinstance(command, Asc2Req) or isinstance(command, Stm2Req):
                    self._state = Switch.THRU if command.is_thru else Switch.OUT
                else:
                    log.warning(f"Unhandled Switch State Update received: {command}")
                self.changed.set()
                self.synchronizer.notify_all()
            # inform the routes that include this switch of new state
            self.update_route_state()

    @property
    def state(self) -> Switch:
        return self._state

    @property
    def is_known(self) -> bool:
        return self._state is not None

    @property
    def is_thru(self) -> bool:
        return self._state == Switch.THRU

    @property
    def is_through(self) -> bool:
        return self.is_thru

    @property
    def is_out(self) -> bool:
        return self._state == Switch.OUT

    @property
    def payload(self) -> str:
        return f"{self._state.name if self._state is not None else 'Unknown'}"

    def register_route(self, route: RouteState) -> None:
        self._routes.add(route)

    def update_route_state(self) -> None:
        for route in self._routes:
            route.update_switch_state(self)

    def as_bytes(self) -> bytes:
        if self.comp_data is None:
            self.initialize(self.scope, self.address)
        byte_str = super().as_bytes()
        if self.is_known:
            byte_str += CommandReq.build(self.state, self.address).as_bytes
        return byte_str

    def as_dict(self) -> Dict[str, Any]:
        d = super()._as_dict()
        if self.is_known:
            state = "thru" if self.is_through else "out"
        else:
            state = None
        d["state"] = state
        return d


class RouteState(TmccState):
    """
    Maintain Route State
    """

    def __init__(self, scope: CommandScope = CommandScope.ROUTE) -> None:
        if scope != CommandScope.ROUTE:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._signature: dict[int, bool] = dict()
        self._current_state: dict[int, bool | None] = dict()

    def update(self, command: L | P) -> None:
        if command:
            if command.command == TMCC1HaltCommandEnum.HALT:
                return
            with self.synchronizer:
                if not self.is_comp_data_record:
                    if isinstance(command, CommandReq):
                        from ..comm.command_listener import CommandDispatcher

                        log.info(f"Still awaiting for initial state, will retry {command}...")
                        CommandDispatcher.get().offer(command)
                        return
                super().update(command)
                if isinstance(command, CompDataMixin) and command.is_comp_data_record:
                    self._update_comp_data(command.comp_data)
                    # set up callbacks so that changes to component switch states
                    # can real-time trigger updates to this route's state
                    comps = self.components
                    if comps:
                        from .component_state_store import ComponentStateStore

                        store = ComponentStateStore.get()
                        for comp in comps:
                            self._signature.update(comp.as_signature)
                            switch = store.get_state(CommandScope.SWITCH, comp.tmcc_id, True)
                            if isinstance(switch, SwitchState):
                                self._current_state.update(
                                    {switch.address: switch.is_thru if switch.is_known else None}
                                )
                                switch.register_route(self)
                elif isinstance(command, CommandReq):
                    pass
                else:
                    log.warning(f"Unhandled Route State Update received: {command}")
                self.changed.set()
                self._cv.notify_all()

    @property
    def components(self) -> List[RouteComponent] | None:
        return self.comp_data.components.copy() if self.comp_data.components else None

    @property
    def payload(self) -> str:
        pl = f"Active: {'True ' if self.is_active else 'False'}"
        pl += f" {self.comp_data.payload()}" if self.comp_data else ""
        return pl

    @property
    def is_active(self) -> bool:
        return self._signature == self._current_state

    @property
    def as_signature(self) -> dict[int, bool]:
        return self._signature

    def update_switch_state(self, switch: SwitchState) -> None:
        with self.synchronizer:
            self._current_state.update({switch.address: switch.is_thru if switch.is_known else None})
            self.changed.set()
            self._cv.notify_all()

    def as_dict(self) -> Dict[str, Any]:
        d = super()._as_dict()
        d["active"] = self.is_active
        if self.components:
            sw = [{"switch": c.tmcc_id, "position": "thru" if c.is_thru is True else "out"} for c in self.components]
        else:
            sw = list()
        d["switches"] = sw
        return d


T = TypeVar("T", bound=ComponentState)

# noinspection PyTypeChecker,PyTypeHints
SCOPE_TO_STATE_MAP: dict[CommandScope, T] = {
    CommandScope.ROUTE: RouteState,
    CommandScope.SWITCH: SwitchState,
}


class ThreadSafeDefaultDict(defaultdict):
    def __init__(self) -> None:
        super().__init__(None)
        self._lock = threading.RLock()

    def __getitem__(self, key):
        with self._lock:
            return super().__getitem__(key)

    def __setitem__(self, key, value: Any) -> None:
        with self._lock:
            super().__setitem__(key, value)

    def __delitem__(self, key) -> None:
        with self._lock:
            super().__delitem__(key)

    def __len__(self) -> int:
        with self._lock:
            return super().__len__()

    def __contains__(self, key):
        with self._lock:
            return super().__contains__(key)

    def get(self, key, default=None) -> Any:
        with self._lock:
            return super().get(key, default)


class SystemStateDict(ThreadSafeDefaultDict):
    """
    Maintains a dictionary of CommandScope to ComponentStateDict
    """

    def __missing__(self, key: CommandScope | tuple[CommandScope, int]) -> ComponentStateDict:
        """
        generate a ComponentState object for the dictionary, based on the key
        """
        if isinstance(key, CommandScope) and key in SCOPE_TO_STATE_MAP:
            scope = key
        else:
            raise KeyError(f"Invalid scope key: {key}")
        # create the component state dict for this key
        with self._lock:
            self[key] = ComponentStateDict(scope)
            return self[key]


class ComponentStateDict(ThreadSafeDefaultDict):
    from .accessory_state import AccessoryState  # noqa: F401
    from .base_state import BaseState  # noqa: F401
    from .block_state import BlockState  # noqa: F401
    from .engine_state import EngineState, TrainState  # noqa: F401
    from .irda_state import IrdaState  # noqa: F401
    from .sync_state import SyncState  # noqa: F401

    def __init__(self, scope: CommandScope):
        super().__init__()  # base class doesn't get a factory
        if scope not in SCOPE_TO_STATE_MAP:
            raise ValueError(f"Invalid scope: {scope}")
        self._scope = scope

    @property
    def scope(self) -> CommandScope:
        return self._scope

    # noinspection PyCallingNonCallable
    def __missing__(self, key: int) -> ComponentState:
        """
        generate a ComponentState object for the dictionary, based on the key
        """
        if not isinstance(key, int):
            raise KeyError(f"Invalid ID: {key}")
        elif self.scope == CommandScope.BASE and key != 0:
            raise KeyError(f"Invalid ID: {key}")
        elif self.scope == CommandScope.SYNC and key != 99:
            raise KeyError(f"Invalid ID: {key}")
        elif self.scope == CommandScope.ENGINE and (key < 1 or key > 9999):
            raise KeyError(f"Invalid ID: {key}")
        elif self.scope not in {CommandScope.BASE, CommandScope.ENGINE, CommandScope.SYNC} and (key < 1 or key > 99):
            raise KeyError(f"Invalid ID: {key}")
        with self._lock:
            value: ComponentState = SCOPE_TO_STATE_MAP[self._scope](self._scope)
            value._address = key
            self[key] = value
            return self[key]



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

from __future__ import annotations

from dataclasses import dataclass

from ..db.component_state import L, T
from ..protocol.constants import CommandScope, Mixins
from ..utils.singleton import singleton
from ..utils.validations import Validations
from .constants import PDI_EOP, PDI_SOP, Amc2Action, PdiCommand
from .lcs_req import LcsReq

NULL_BYTE = b"\x00"


class AccessType(Mixins):
    ENGINE = 0
    TRAIN = 1
    ACC = 2


class OutputType(Mixins):
    NORMAL = 0
    DELTA = 1
    AC = 2

    @property
    def is_dc(self) -> bool:
        return self in {OutputType.DELTA, OutputType.NORMAL}

    @property
    def is_ac(self) -> bool:
        return self in {OutputType.AC}

    @property
    def label(self) -> str:
        if self == OutputType.AC:
            return OutputType.AC.name
        return self.name.capitalize()


class Direction(Mixins):
    FORWARD = 1
    REVERSE = 2
    AC = 3

    @property
    def label(self) -> str:
        if self == Direction.AC:
            return Direction.AC.name
        return self.name.capitalize()


@dataclass(slots=True)
class Amc2Motor:
    id: int
    output_type: OutputType
    direction: Direction
    restore: bool
    restore_state: bool
    speed: int

    def __repr__(self) -> str:
        t = f"Output: {self.output_type.label}"
        d = f"Dir: {self.direction.label}"
        r = f"Restore: {self.restore}"
        if self.restore:
            r += f" to {'On' if self.restore_state else 'Off'}"
        return f"Mo #{self.id} {t} {d} Speed: {self.speed} {r}"


@dataclass(slots=True)
class Amc2Lamp:
    id: int
    level: int

    def __repr__(self) -> str:
        return f"Lm #{self.id} level: {self.level}"


class Amc2Req(LcsReq):
    # noinspection PyUnusedLocal,PyTypeChecker
    @classmethod
    def request_config(cls, state: T, cmd: L) -> Amc2Req:
        return cls(state.address, pdi_command=PdiCommand.AMC2_GET, action=Amc2Action.CONFIG)

    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.AMC2_GET,
        action: Amc2Action = Amc2Action.CONFIG,
        ident: int | None = None,
        error: bool = False,
        debug: int = None,
        motor: int | None = None,
        speed: int = None,
        direction: Direction | None = None,
        output_type: OutputType | None = None,
        restore_state: bool | None = None,
        lamp: int | None = None,
        level: int = None,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        self.scope = CommandScope.ACC
        if isinstance(data, bytes):
            # initialize all
            self._debug = False
            self._motor1 = self._motor2 = None
            self._lamp1 = self._lamp2 = self._lamp3 = self._lamp4 = None
            self._option = None
            self._access_type = None
            self._motor = self._speed = self._direction = self._lamp = self._level = None
            self._output_type = self._restore_state = None

            # what is the request type?
            self._action = Amc2Action(self._action_byte)
            data_len = len(self._data)
            if self._action == Amc2Action.CONFIG:
                self._debug = self._data[4] if data_len > 4 else None
                self._option = self._data[5:7] if data_len > 6 else None
                self._access_type = AccessType(self._data[7]) if data_len > 7 else None
                self._motor1, self._motor2 = self._harvest_motors(self._data[8:18]) if data_len > 17 else (None, None)
                self._lamp1 = Amc2Lamp(1, self._data[18]) if data_len > 18 else None
                self._lamp2 = Amc2Lamp(2, self._data[19]) if data_len > 19 else None
                self._lamp3 = Amc2Lamp(3, self._data[20]) if data_len > 20 else None
                self._lamp4 = Amc2Lamp(4, self._data[21]) if data_len > 21 else None
            elif self._action == Amc2Action.MOTOR:
                self._motor = self._data[3] if data_len > 3 else None
                self._speed = self._data[4] if data_len > 4 else None
                self._direction = self._data[5] if data_len > 5 else None
            elif self._action == Amc2Action.LAMP:
                self._lamp = self._data[3] if data_len > 3 else None
                self._level = self._data[4] if data_len > 4 else None
            elif self._action == Amc2Action.MOTOR_CONFIG:
                self._motor = self._data[3] if data_len > 3 else None
                self._output_type = self._data[4] if data_len > 4 else None
                self._restore_state = bool(self._data[5]) if data_len > 5 else None
        else:
            self._action = action
            self._debug = debug
            self._motor = Validations.validate_int(motor, 0, 1, "Motor", True)
            self._speed = Validations.validate_int(speed, 0, 100, "Speed", True)
            self._direction = direction.value if direction else Direction.AC.value
            self._output_type = output_type.value if output_type else OutputType.AC
            self._restore_state = restore_state if restore_state is not None else True
            self._lamp = Validations.validate_int(lamp, 0, 3, "Lamp", True)
            self._level = Validations.validate_int(level, 0, 100, "Level", True)

    @property
    def debug(self) -> int | None:
        return self._debug

    @property
    def access_type(self) -> AccessType:
        return self._access_type

    @property
    def motor(self) -> int:
        return self._motor

    @property
    def speed(self) -> int:
        return self._speed

    @property
    def direction(self) -> Direction:
        return Direction(self._direction) if self._direction is not None else Direction.AC

    @property
    def output_type(self) -> OutputType:
        return OutputType(self._output_type) if self._output_type is not None else OutputType.AC

    @property
    def restore_state(self) -> bool:
        return self._restore_state if self._restore_state is not None else True

    @property
    def motor1(self) -> Amc2Motor:
        return self._motor1

    @property
    def motor2(self) -> Amc2Motor:
        return self._motor2

    @property
    def lamp(self) -> int:
        return self._lamp

    @property
    def level(self) -> int:
        return self._level

    @property
    def lamp1(self) -> Amc2Lamp:
        return self._lamp1

    @property
    def lamp2(self) -> Amc2Lamp:
        return self._lamp2

    @property
    def lamp3(self) -> Amc2Lamp:
        return self._lamp3

    @property
    def lamp4(self) -> Amc2Lamp:
        return self._lamp4

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        if self.pdi_command != PdiCommand.AMC2_GET:
            if self.action == Amc2Action.CONFIG:
                at = f"Type: {self._access_type.label}"
                m1 = f"{self._motor1}"
                m2 = f"{self._motor2}"
                l1 = f"{self._lamp1}"
                return f"{at} {m1} {m2} {l1} Debug: {self.debug} ({self.packet})"
            elif self._action == Amc2Action.MOTOR:
                return f"Motor: {self.motor + 1} Speed: {self.speed} Dir: {self.direction.label} ({self.packet})"
            elif self._action == Amc2Action.LAMP:
                return f"Lamp: {self.lamp + 1} Level: {self.level} ({self.packet})"
            elif self._action == Amc2Action.MOTOR_CONFIG:
                ot = f"Output Type: {self.output_type.label}"
                rs = f"Restore State: {'On' if self._restore_state else 'Off'}"
                return f"Motor: {self.motor + 1} {ot} {rs} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += self.action.as_bytes
        if self._action == Amc2Action.CONFIG:
            if self.pdi_command != PdiCommand.AMC2_GET:
                byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
                debug = self.debug if self.debug is not None else 0
                byte_str += debug.to_bytes(1, byteorder="big")
                byte_str += self._option if self._option else (0x0000).to_bytes(2, byteorder="big")
                byte_str += self._access_type.value.to_bytes(1, byteorder="big") if self._access_type else NULL_BYTE
                byte_str += self._motors_as_bytes()
                for lamp in [self._lamp1, self._lamp2, self._lamp3, self._lamp4]:
                    byte_str += lamp.level.to_bytes(1, byteorder="big") if lamp else NULL_BYTE
        elif self._action == Amc2Action.MOTOR:
            byte_str += self.motor.to_bytes(1, byteorder="big")
            if self.pdi_command != PdiCommand.AMC2_GET:
                byte_str += self.speed.to_bytes(1, byteorder="big")
                byte_str += self._direction.to_bytes(1, byteorder="big")
        elif self._action == Amc2Action.LAMP:
            byte_str += self.lamp.to_bytes(1, byteorder="big")
            if self.pdi_command != PdiCommand.AMC2_GET:
                byte_str += self.level.to_bytes(1, byteorder="big")
        elif self._action == Amc2Action.MOTOR_CONFIG:
            byte_str += self.motor.to_bytes(1, byteorder="big")
            if self.pdi_command != PdiCommand.AMC2_GET:
                byte_str += self.output_type.value.to_bytes(1, byteorder="big")
                byte_str += self.restore_state.to_bytes(1, byteorder="big")
        elif self._action == Amc2Action.IDENTIFY:
            if self.pdi_command == PdiCommand.AMC2_SET:
                byte_str += (self.ident if self.ident is not None else 0).to_bytes(1, byteorder="big")
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @staticmethod
    def _harvest_motors(data: bytes) -> tuple[Amc2Motor, Amc2Motor]:
        motor1 = Amc2Motor(
            1,
            OutputType(data[0]),
            Direction(data[2]),
            bool(data[4]),
            bool(data[6]),
            data[8],
        )
        motor2 = Amc2Motor(
            2,
            OutputType(data[1]),
            Direction(data[3]),
            bool(data[5]),
            bool(data[7]),
            data[9],
        )
        return motor1, motor2

    def _motors_as_bytes(self) -> bytes:
        byte_str = bytes()
        byte_str += self._motor1.output_type.value.to_bytes(1, byteorder="big") if self._motor1 else NULL_BYTE
        byte_str += self._motor2.output_type.value.to_bytes(1, byteorder="big") if self._motor2 else NULL_BYTE

        byte_str += self._motor1.direction.value.to_bytes(1, byteorder="big") if self._motor1 else NULL_BYTE
        byte_str += self._motor2.direction.value.to_bytes(1, byteorder="big") if self._motor2 else NULL_BYTE

        byte_str += self._motor1.restore.to_bytes(1, byteorder="big") if self._motor1 else NULL_BYTE
        byte_str += self._motor2.restore.to_bytes(1, byteorder="big") if self._motor2 else NULL_BYTE

        byte_str += self._motor1.restore_state.to_bytes(1, byteorder="big") if self._motor1 else NULL_BYTE
        byte_str += self._motor2.restore_state.to_bytes(1, byteorder="big") if self._motor2 else NULL_BYTE

        byte_str += self._motor1.speed.to_bytes(1, byteorder="big") if self._motor1 else NULL_BYTE
        byte_str += self._motor2.speed.to_bytes(1, byteorder="big") if self._motor2 else NULL_BYTE
        return byte_str


@singleton
class Amc2StateSync:
    from .pdi_listener import PdiListener

    def __init__(self, listener: PdiListener) -> None:
        self._pdi_listener = listener
        self._pdi_dispatcher = dispatcher = listener.dispatcher
        dispatcher.subscribe(self, PdiCommand.AMC2_RX, action=Amc2Action.MOTOR)
        dispatcher.subscribe(self, PdiCommand.AMC2_RX, action=Amc2Action.LAMP)
        dispatcher.subscribe(self, PdiCommand.AMC2_RX, action=Amc2Action.MOTOR_CONFIG)

    def __call__(self, command: Amc2Req) -> None:
        # if we receive a motor/lamp/motor_config RX, request complete configuration
        if isinstance(command, Amc2Req) and command.action in {
            Amc2Action.LAMP,
            Amc2Action.MOTOR,
            Amc2Action.MOTOR_CONFIG,
        }:
            # send config command to retrieve complete config
            Amc2Req(command.tmcc_id, PdiCommand.AMC2_GET, Amc2Action.CONFIG).send()



from threading import Thread, Event
from typing import List

from RPLCD.i2c import CharLCD
import sys

sys.setswitchinterval(0.001)

SMILEY_CHAR = (
    0b00000,
    0b01010,
    0b01010,
    0b00000,
    0b10001,
    0b10001,
    0b01110,
    0b00000,
)

LCD_PCF8574_ADDRESS = 0x27


class Lcd(CharLCD):
    def __init__(
        self,
        i2c_expander: str = "PCF8574",
        address: int = LCD_PCF8574_ADDRESS,
        port: int = 1,
        rows: int = 4,
        cols: int = 20,
        scroll_speed: float = 0.5,
        dot_size: int = 8,
        char_map: str = "A02",
        auto_linebreaks: bool = False,
        backlight_enabled: bool = True,
    ):
        self._frame_buffer = list()  # need to define before super because super calls clear...
        self._scroller = None  # same here
        super().__init__(
            i2c_expander=i2c_expander,
            address=address,
            port=port,
            cols=cols,
            rows=rows,
            dotsize=dot_size,
            charmap=char_map,
            auto_linebreaks=auto_linebreaks,
            backlight_enabled=backlight_enabled,
        )
        self._rows = rows
        self._cols = cols
        self._scroll_speed = scroll_speed
        self._backlight_enabled = backlight_enabled
        self._auto_linebreaks = auto_linebreaks
        self.cursor_mode = "hide"
        self.clear()

    @property
    def rows(self) -> int:
        return self._rows

    @property
    def cols(self) -> int:
        return self._cols

    @property
    def frame_buffer(self) -> List[str]:
        return self._frame_buffer

    def add(self, row: str) -> bool:
        if len(self._frame_buffer) < self.rows:
            self._frame_buffer.append(row)
            return True
        return False

    def clear(self) -> None:
        super().clear()
        self.home()
        self.clear_frame_buffer()
        self.cursor_mode = "hide"

    def clear_frame_buffer(self) -> None:
        self.stop_scrolling()
        self._frame_buffer.clear()

    def stop_scrolling(self) -> None:
        if self._scroller is not None:
            """
            Shutdown the scrolling thread and wait for it to finish,
            otherwise, we could have multiple threads trying to
            output to the screen at the same time.
            """
            self._scroller.shutdown()
            self._scroller.join()
            self._scroller = None

    def write_frame_buffer(self, clear_display: bool = True) -> None:
        self.stop_scrolling()
        if clear_display is True:
            super().clear()  # call the super, otherwise frame buffer is cleared
        self.home()  # reposition cursor
        for r, row in enumerate(self._frame_buffer):
            if row:
                self.cursor_pos = (r, 0)
                self.write_string(row.ljust(self.cols)[: self.cols])
        if (
            self._scroll_speed > 0.0
            and len(self._frame_buffer) > 0
            and self._frame_buffer[0]
            and len(self._frame_buffer[0]) > self.cols
        ):
            self._scroller = Scroller(self, self._frame_buffer[0], self._scroll_speed)
        self.cursor_mode = "hide"

    def print(self, c: int | str) -> None:
        if isinstance(c, int) and 0 <= c <= 255:
            self.write(c)
        elif isinstance(c, str) and len(c) == 1:
            self.write(ord(c[0]))
        elif isinstance(c, str):
            self.write_string(c)
        else:
            self.write_string(str(c))


class Scroller(Thread):
    def __init__(self, lcd: Lcd, buffer: str, scroll_speed: float = 0.5) -> None:
        super().__init__(daemon=True)
        self._lcd = lcd
        self._buffer = buffer.strip()
        self._scroll_speed = scroll_speed
        self._exit = Event()
        self.start()

    def shutdown(self) -> None:
        self._exit.set()

    def run(self) -> None:
        s = self._buffer + " " + self._buffer
        try:
            while not self._exit.is_set():
                for i in range(len(self._buffer) + 1):
                    self._lcd.cursor_pos = (0, 0)
                    self._lcd.write_string(s[i : i + self._lcd.cols])
                    self._exit.wait(self._scroll_speed)
                    if self._exit.is_set():
                        return
        finally:
            self._lcd.cursor_pos = (0, 0)
            self._lcd.cursor_mode = "hide"



# noinspection PyPackageRequirements
import pytest

from src.pytrain.protocol.multibyte.multibyte_constants import *
from src.pytrain.protocol.tmcc1.tmcc1_constants import *
from src.pytrain.protocol.tmcc2.tmcc2_constants import *
from ..test_base import TestBase


class TestConstants(TestBase):
    def test_by_name_mixin(self) -> None:
        # assert all enums are found
        for ss in TMCC1SwitchCommandEnum:
            assert TMCC1SwitchCommandEnum.by_name(ss.name) == ss

        # assert by_name is case-insensitive
        assert TMCC1SwitchCommandEnum.by_name("thru") == TMCC1SwitchCommandEnum.THRU
        assert TMCC1SwitchCommandEnum.by_name("THRU") == TMCC1SwitchCommandEnum.THRU

        # assert non-members return None
        assert TMCC1SwitchCommandEnum.by_name("NOT_PRESENT") is None

        # assert None return None
        assert TMCC1SwitchCommandEnum.by_name(str(None)) is None

        # check ValueError is thrown
        with pytest.raises(ValueError, match="'NOT_PRESENT' is not a valid TMCC1SwitchCommandEnum"):
            TMCC1SwitchCommandEnum.by_name("NOT_PRESENT", raise_exception=True)

        # check ValueError is thrown
        with pytest.raises(ValueError, match="None is not a valid TMCC1SwitchCommandEnum"):
            TMCC1SwitchCommandEnum.by_name(None, raise_exception=True)  # noqa

        # check ValueError is thrown
        with pytest.raises(ValueError, match="Empty is not a valid TMCC1SwitchCommandEnum"):
            TMCC1SwitchCommandEnum.by_name("  ", raise_exception=True)

    def test_by_name_mixin_in_enums(self) -> None:
        """
        Test that all defined enums have Mixins
        """
        enums = [
            CommandSyntax,
            CommandScope,
            TMCC1SwitchCommandEnum,
            TMCC1HaltCommandEnum,
            TMCC1RouteCommandEnum,
            TMCC1AuxCommandEnum,
            TMCC1EngineCommandEnum,
            TMCC2ParameterIndex,
            TMCC2EngineCommandEnum,
            TMCC2EffectsControl,
            TMCC2LightingControl,
        ]
        for env in enums:
            for en in env:
                assert env.by_name(en.name) == en

    def test_tmcc1_constants(self) -> None:
        """
        All bit patterns are from the Lionel LCS Partner Documentation,
        Legacy Command Protocol, rev 1.21
        """
        assert TMCC1_COMMAND_PREFIX == 0b11111110

        assert TMCC1_HALT_COMMAND == 0b1111111111111111

        assert TMCC1_ROUTE_COMMAND == 0b1101000000011111

        assert TMCC1_SWITCH_THRU_COMMAND == 0b0100000000000000
        assert TMCC1_SWITCH_OUT_COMMAND == 0b0100000000011111
        assert TMCC1_SWITCH_SET_ADDRESS_COMMAND == 0b0100000000101011

        # test tmcc1 acc commands
        acc_cmd_prefix = 0x8000
        assert TMCC1_ACC_ON_COMMAND == acc_cmd_prefix | 0b0101111
        assert TMCC1_ACC_OFF_COMMAND == acc_cmd_prefix | 0b0100000
        assert TMCC1_ACC_NUMERIC_COMMAND == acc_cmd_prefix | 0b0010000
        assert TMCC1_ACC_SET_ADDRESS_COMMAND == acc_cmd_prefix | 0b0101011

        assert TMCC1_ACC_AUX_2_OFF_COMMAND == acc_cmd_prefix | 0b0001100
        assert TMCC1_ACC_AUX_2_OPTION_1_COMMAND == acc_cmd_prefix | 0b0001101
        assert TMCC1_ACC_AUX_2_OPTION_2_COMMAND == acc_cmd_prefix | 0b0001110
        assert TMCC1_ACC_AUX_2_ON_COMMAND == acc_cmd_prefix | 0b0001111

        assert TMCC1_ACC_AUX_1_OFF_COMMAND == acc_cmd_prefix | 0b0001000
        assert TMCC1_ACC_AUX_1_OPTION_1_COMMAND == acc_cmd_prefix | 0b0001001
        assert TMCC1_ACC_AUX_1_OPTION_2_COMMAND == acc_cmd_prefix | 0b0001010
        assert TMCC1_ACC_AUX_1_ON_COMMAND == acc_cmd_prefix | 0b0001011

        # test tmcc1 engine/train commands
        eng_cmd_prefix = 0x0000
        assert TMCC1_ENG_ABSOLUTE_SPEED_COMMAND == eng_cmd_prefix | 0b1100000
        assert TMCC1_ENG_RELATIVE_SPEED_COMMAND == eng_cmd_prefix | 0b1000000
        assert TMCC1_ENG_FORWARD_DIRECTION_COMMAND == eng_cmd_prefix | 0b0000000
        assert TMCC1_ENG_TOGGLE_DIRECTION_COMMAND == eng_cmd_prefix | 0b0000001
        assert TMCC1_ENG_REVERSE_DIRECTION_COMMAND == eng_cmd_prefix | 0b0000011
        assert TMCC1_ENG_BOOST_SPEED_COMMAND == eng_cmd_prefix | 0b0000100
        assert TMCC1_ENG_BRAKE_SPEED_COMMAND == eng_cmd_prefix | 0b0000111
        assert TMCC1_ENG_OPEN_FRONT_COUPLER_COMMAND == eng_cmd_prefix | 0b0000101
        assert TMCC1_ENG_OPEN_REAR_COUPLER_COMMAND == eng_cmd_prefix | 0b0000110
        assert TMCC1_ENG_BLOW_HORN_ONE_COMMAND == eng_cmd_prefix | 0b0011100
        assert TMCC1_ENG_RING_BELL_COMMAND == eng_cmd_prefix | 0b0011101
        assert TMCC1_ENG_LET_OFF_SOUND_COMMAND == eng_cmd_prefix | 0b0011110
        assert TMCC1_ENG_BLOW_HORN_TWO_COMMAND == eng_cmd_prefix | 0b0011111

        assert TMCC1_ENG_AUX1_OFF_COMMAND == eng_cmd_prefix | 0b0001000
        assert TMCC1_ENG_AUX1_OPTION_ONE_COMMAND == eng_cmd_prefix | 0b0001001
        assert TMCC1_ENG_AUX1_OPTION_TWO_COMMAND == eng_cmd_prefix | 0b0001010
        assert TMCC1_ENG_AUX1_ON_COMMAND == eng_cmd_prefix | 0b0001011

        assert TMCC1_ENG_AUX2_OFF_COMMAND == eng_cmd_prefix | 0b0001100
        assert TMCC1_ENG_AUX2_OPTION_ONE_COMMAND == eng_cmd_prefix | 0b0001101
        assert TMCC1_ENG_AUX2_OPTION_TWO_COMMAND == eng_cmd_prefix | 0b0001110
        assert TMCC1_ENG_AUX2_ON_COMMAND == eng_cmd_prefix | 0b0001111

        assert TMCC1_ENG_SET_MOMENTUM_LOW_COMMAND == eng_cmd_prefix | 0b0101000
        assert TMCC1_ENG_SET_MOMENTUM_MEDIUM_COMMAND == eng_cmd_prefix | 0b0101001
        assert TMCC1_ENG_SET_MOMENTUM_HIGH_COMMAND == eng_cmd_prefix | 0b0101010

        assert TMCC1_ENG_NUMERIC_COMMAND == eng_cmd_prefix | 0b0010000

        assert TMCC1_ENG_SET_ADDRESS_COMMAND == eng_cmd_prefix | 0b0101011

    def test_tmcc2_constants(self) -> None:
        """
        All bit patterns are from the Lionel LCS Partner Documentation,
        Legacy Command Protocol, rev 1.21
        """
        assert TMCC2_AUX1_OFF_COMMAND == 0b100001000
        assert TMCC2_AUX1_ON_COMMAND == 0b100001011
        assert TMCC2_AUX1_OPTION_ONE_COMMAND == 0b100001001
        assert TMCC2_AUX1_OPTION_TWO_COMMAND == 0b100001010
        assert TMCC2_AUX2_OFF_COMMAND == 0b100001100
        assert TMCC2_AUX2_ON_COMMAND == 0b100001111
        assert TMCC2_AUX2_OPTION_ONE_COMMAND == 0b100001101
        assert TMCC2_AUX2_OPTION_TWO_COMMAND == 0b100001110
        assert TMCC2_BELL_OFF_COMMAND == 0b111110100
        assert TMCC2_BELL_ONE_SHOT_DING_COMMAND == 0b111110000
        assert TMCC2_BELL_ON_COMMAND == 0b111110101
        assert TMCC2_BELL_SLIDER_POSITION_COMMAND == 0b110110000
        assert TMCC2_BLOW_HORN_ONE_COMMAND == 0b100011100
        assert TMCC2_BLOW_HORN_TWO_COMMAND == 0b100011111
        assert TMCC2_BOOST_SPEED_COMMAND == 0b100000100
        assert TMCC2_BRAKE_SPEED_COMMAND == 0b100000111
        assert TMCC2_DIESEL_RPM_SOUND_COMMAND == 0b110100000
        assert TMCC2_ENGINE_LABOR_COMMAND == 0b111000000
        assert TMCC2_ENG_AUGER_SOUND_COMMAND == 0b111110111
        assert TMCC2_ENG_BRAKE_AIR_RELEASE_SOUND_COMMAND == 0b111111000
        assert TMCC2_ENG_BRAKE_SQUEAL_SOUND_COMMAND == 0b111110110
        assert TMCC2_ENG_LET_OFF_LONG_SOUND_COMMAND == 0b111111010
        assert TMCC2_ENG_LET_OFF_SOUND_COMMAND == 0b111111001
        assert TMCC2_ENG_REFUELLING_SOUND_COMMAND == 0b100101101
        assert TMCC2_FORWARD_DIRECTION_COMMAND == 0b100000000
        assert TMCC2_HALT_COMMAND == 0b110101011
        assert TMCC2_NUMERIC_COMMAND == 0b100010000
        assert TMCC2_OPEN_FRONT_COUPLER_COMMAND == 0b100000101
        assert TMCC2_OPEN_REAR_COUPLER_COMMAND == 0b100000110
        assert TMCC2_QUILLING_HORN_COMMAND == 0b111100000
        assert TMCC2_REVERSE_DIRECTION_COMMAND == 0b100000011
        assert TMCC2_RING_BELL_COMMAND == 0b100011101
        assert TMCC2_SET_ABSOLUTE_SPEED_COMMAND == 0b000000000
        assert TMCC2_SET_ADDRESS_COMMAND == 0b100101011
        assert TMCC2_SET_BOOST_LEVEL_COMMAND == 0b011101000
        assert TMCC2_SET_BRAKE_LEVEL_COMMAND == 0b011100000
        assert TMCC2_SET_MOMENTUM_COMMAND == 0b011001000
        assert TMCC2_SET_MOMENTUM_HIGH_COMMAND == 0b100101010
        assert TMCC2_SET_MOMENTUM_LOW_COMMAND == 0b100101000
        assert TMCC2_SET_MOMENTUM_MEDIUM_COMMAND == 0b100101001
        assert TMCC2_SET_RELATIVE_SPEED_COMMAND == 0b101000000
        assert TMCC2_SET_TRAIN_BRAKE_COMMAND == 0b011110000
        assert TMCC2_SHUTDOWN_SEQ_ONE_COMMAND == 0b111111101
        assert TMCC2_SHUTDOWN_SEQ_TWO_COMMAND == 0b111111110
        assert TMCC2_SOUND_OFF_COMMAND == 0b101010000
        assert TMCC2_SOUND_ON_COMMAND == 0b101010001
        assert TMCC2_STALL_COMMAND == 0b011111000
        assert TMCC2_START_UP_SEQ_ONE_COMMAND == 0b111111011
        assert TMCC2_START_UP_SEQ_TWO_COMMAND == 0b111111100
        assert TMCC2_STOP_IMMEDIATE_COMMAND == 0b011111011
        assert TMCC2_TOGGLE_DIRECTION_COMMAND == 0b100000001
        assert TMCC2_WATER_INJECTOR_SOUND_COMMAND == 0b110101000
        assert TMCC2_MOTION_START_COMMAND == 0b011111010
        assert TMCC2_MOTION_STOP_COMMAND == 0b011111110
        assert TMCC2_ENG_CYLINDER_HISS_SOUND_COMMAND == 0b101010010
        assert TMCC2_ENG_POP_OFF_SOUND_COMMAND == 0b101010011

    def test_command_scope_enum(self) -> None:
        # check that engine and train elements are in TMCC1CommandIdentifier
        assert TMCC1CommandIdentifier.ENGINE == TMCC1CommandIdentifier(CommandScope.ENGINE.name)
        assert TMCC1CommandIdentifier.TRAIN == TMCC1CommandIdentifier(CommandScope.TRAIN.name)

        # check that engine and train elements are in TMCC2CommandPrefix
        assert TMCC2CommandPrefix.ENGINE == TMCC2CommandPrefix(CommandScope.ENGINE.name)
        assert TMCC2CommandPrefix.TRAIN == TMCC2CommandPrefix(CommandScope.TRAIN.name)

    def test_engine_option_enum(self) -> None:
        # should contain no elements
        assert len(CommandDefEnum) == 0

        # validate _missing_ method is present and throws exception
        with pytest.raises(ValueError, match="FOO is not a valid CommandDefEnum"):
            CommandDefEnum._missing_("foo")

    def test_engine_option(self) -> None:
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2014-20 Richard Hull and contributors
# See LICENSE.rst for details.

import luma.oled.const as const
from luma.core.framebuffer import full_frame
from luma.oled.device.greyscale import greyscale_device

__all__ = ["sh1122"]


# noinspection PyPep8Naming
class ssh1122_const(
    const.common
):  # copying data shit: https://www.displayfuture.com/Display/datasheet/controller/SH1122.pdf
    SET_COL_ADR_LSB = 0x0
    SET_COL_ADR_MSB = 0x10
    SET_DISP_START_LINE = 0x40
    SET_CONTRAST = 0x81
    SET_SEG_REMAP = 0xA0
    SET_ENTIRE_ON = 0xA4
    SET_ENTIRE_OFF = 0xA5
    SET_NORM_INV = 0xA6
    SET_MUX_RATIO = 0xA8
    SET_CTRL_DCDC = 0xAD
    SET_DISP = 0xAE
    SET_ROW_ADR = 0xB0
    SET_COM_OUT_DIR = 0xC0
    SET_DISP_OFFSET = 0xD3
    SET_DISP_CLK_DIV = 0xD5
    SET_PRECHARGE = 0xD9
    SET_VCOM_DESEL = 0xDB
    SET_VSEG_LEVEL = 0xDC
    SET_DISCHARGE_LEVEL = 0x30


# noinspection PyPep8Naming
class sh1122(greyscale_device):
    def __init__(self, serial_interface=None, **kwargs):
        super().__init__(
            ssh1122_const,
            serial_interface,
            256,
            64,
            0,
            "RGB",
            full_frame(),
            nibble_order=0,
            **kwargs,
        )

    def command(self, *args):
        return super().command(*args)

    def _supported_dimensions(self):
        return [(256, 64)]  # I don't know about anything other...

    def _init_sequence(self):
        for cmd in (
            self._const.SET_DISP | 0x00,  # off
            # address setting
            self._const.SET_COL_ADR_LSB,
            self._const.SET_COL_ADR_MSB,  # horizontal
            self._const.SET_ROW_ADR,
            0,
            # resolution and layout
            self._const.SET_DISP_START_LINE | 0x00,
            self._const.SET_SEG_REMAP,
            self._const.SET_MUX_RATIO,
            self.height - 1,
            self._const.SET_COM_OUT_DIR,  # scan from COM0 to COM[N]
            self._const.SET_DISP_OFFSET,
            0x00,
            0b11010101,
            0b11110000,
            # display
            self._const.SET_CONTRAST,
            0x80,  # median
            self._const.SET_ENTIRE_ON,  # output follows RAM contents
            self._const.SET_NORM_INV,  # not inverted
            self._const.SET_DISP | 0x01,
        ):
            self.command(cmd)

    def _set_position(self, top, right, bottom, left):
        pass

    def display(self, image):
        """
        Takes a 1-bit monochrome or 24-bit RGB image and renders it
        to the greyscale OLED display. RGB pixels are converted to 8-bit
        greyscale values using a simplified Luma calculation, based on
        *Y=0.299R+0.587G+0.114B*.
        :param image: the image to render
        :type image: PIL.Image.Image
        """

        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        for image, bounding_box in self.framebuffer.redraw(image):
            left, top, right, bottom = bounding_box
            width = right - left
            height = bottom - top

            buf = bytearray(width * height >> 1)

            self._populate(buf, image.getdata())
            self.data(buf)

    def cleanup(self):
        self.command(self._const.SET_ENTIRE_OFF)  # to be absolutely sure
        super().cleanup()  # and this should call DISPLAY OFF command



from __future__ import annotations

import abc
from abc import ABC

from .sequence_constants import SequenceCommandEnum
from .sequence_req import SequenceReq
from ..constants import CommandScope, DEFAULT_ADDRESS
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum
from ...db.component_state_store import ComponentStateStore


class LaborEffectBase(SequenceReq, ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        command: SequenceCommandEnum,
        address: int,
        scope: CommandScope = CommandScope.ENGINE,
        data: int = 0,
        inc: int = 0,
    ) -> None:
        super().__init__(command, address, scope)
        self._inc = inc
        self._scope = scope
        self._data = data
        self._state = None
        self.add(TMCC2EngineCommandEnum.ENGINE_LABOR, scope=CommandScope.ENGINE)

    @property
    def scope(self) -> CommandScope | None:
        return self._scope

    @scope.setter
    def scope(self, new_scope: CommandScope) -> None:
        if new_scope != self._scope:
            # can only change scope for Engine and Train commands, and then, just from the one to the other
            if self._scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                if new_scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                    self._scope = new_scope
                    self._apply_scope()
                    return
            raise AttributeError(f"Scope {new_scope} not supported for {self}")

    def _recalculate(self):
        self._state = ComponentStateStore.get_state(self.scope, self.address, False)
        if self._state:
            labor = self._state.labor + self._inc
            labor = min(max(labor, 0), 31)
            for req_wrapper in self._requests:
                req = req_wrapper.request
                req.data = labor


class LaborEffectUpReq(LaborEffectBase):
    def __init__(
        self,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = CommandScope.ENGINE,
    ) -> None:
        super().__init__(SequenceCommandEnum.LABOR_EFFECT_UP_SEQ, address, scope, data, 1)


SequenceCommandEnum.LABOR_EFFECT_UP_SEQ.value.register_cmd_class(LaborEffectUpReq)


class LaborEffectDownReq(LaborEffectBase):
    def __init__(
        self,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = CommandScope.ENGINE,
    ) -> None:
        super().__init__(SequenceCommandEnum.LABOR_EFFECT_DOWN_SEQ, address, scope, data, -1)


SequenceCommandEnum.LABOR_EFFECT_DOWN_SEQ.value.register_cmd_class(LaborEffectDownReq)



#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
import sys
from abc import ABC, ABCMeta
from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
from typing import Any, List

from ..comm.comm_buffer import CommBuffer
from ..protocol.command_base import CommandBase
from ..protocol.constants import (
    DEFAULT_BAUDRATE,
    DEFAULT_DURATION_INTERVAL_MSEC,
    DEFAULT_PORT,
    DEFAULT_VALID_BAUDRATES,
    PROGRAM_NAME,
    CommandScope,
    CommandSyntax,
)
from ..protocol.tmcc1.tmcc1_constants import TMCC1_SPEED_MAP
from ..protocol.tmcc2.tmcc2_constants import TMCC2_SPEED_MAP
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class CliBase(ABC):
    """
    Base class for PyTrain CLI commands
    """

    __metaclass__ = ABCMeta

    @classmethod
    def command_parser(cls) -> ArgumentParser | None:
        return None

    @staticmethod
    def cli_parser() -> ArgumentParser:
        """
        Add options common to all CLI commands here. Command handlers
        can inherit from this parser to add other command-specific options.
        """
        # define arguments common to all Legacy CLI commands
        parser = PyTrainArgumentParser(add_help=False)

        parser.add_argument(
            "-baudrate",
            action="store",
            type=int,
            choices=DEFAULT_VALID_BAUDRATES,
            default=DEFAULT_BAUDRATE,
            help=f"Baud Rate ({DEFAULT_BAUDRATE})",
        )
        parser.add_argument("-port", action="store", default=DEFAULT_PORT, help=f"Serial Port ({DEFAULT_PORT})")
        parser.add_argument(
            "-server",
            action="store",
            help=f"IP Address of {PROGRAM_NAME} server, if client. Server communicates with Base 3/LCS SER2",
        )
        return parser

    @staticmethod
    def multi_parser() -> ArgumentParser:
        """
        Add options to allow command repetition and delay
        """
        # define arguments common to all Legacy CLI commands
        parser = PyTrainArgumentParser(add_help=False)
        parser.add_argument(
            "-repeat",
            action="store",
            type=CliBase._validate_repeat,
            default=1,
            help="Number of times to repeat command (default: 1)",
        )

        parser.add_argument(
            "-delay",
            action="store",
            type=CliBase._validate_time,
            default=0.0,
            help="Second(s) to delay between repeated commands (default: 0)",
        )

        parser.add_argument(
            "-duration",
            action="store",
            type=CliBase._validate_time,
            default=0.0,
            help="Second(s) to continue firing the command",
        )

        parser.add_argument(
            "-interval",
            action="store",
            type=CliBase._validate_interval,
            default=None,
            help=f"Milliseconds to pause between commands with -duration (default: {DEFAULT_DURATION_INTERVAL_MSEC})",
        )
        return parser

    @staticmethod
    def command_format_parser(default: CommandSyntax = CommandSyntax.LEGACY) -> ArgumentParser:
        """
        Add command_def to run command using TMCC1 command syntax
        """
        parser = PyTrainArgumentParser(add_help=False)
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-tmcc",
            "--tmcc1",
            action="store_const",
            const=CommandSyntax.TMCC,
            dest="format",
            help="Use TMCC1 command syntax.",
        )
        group.add_argument(
            "-legacy",
            "--tmcc2",
            action="store_const",
            const=CommandSyntax.LEGACY,
            dest="format",
            help="Use TMCC2/Legacy command syntax.",
        )
        group.set_defaults(format=default)
        return parser

    @staticmethod
    def train_parser() -> ArgumentParser:
        """
        Add command_def to run command TMCC2 command as train rather than engine
        """
        parser = PyTrainArgumentParser(add_help=False)
        parser.add_argument(
            "-train",
            action="store_const",
            const=True,
            help="Direct command to addressed train rather than engine (for TMCC2 commands)",
        )
        return parser

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        if arg_parser is None:
            arg_parser = self.command_parser()
        if cmd_line is None:
            self._args = arg_parser.parse_args()
        else:
            self._args = arg_parser.parse_args(cmd_line)
        self._command = None
        self._command_format = CommandSyntax.LEGACY  # Use TMCC2-style commands by default, if supported
        self._command_line: List[str] = cmd_line
        self._do_fire = do_fire
        # as the TrainControl cli strips out some commands, we need to restore them
        if "baudrate" in self._args:
            self._baudrate = self._args.baudrate
        else:
            self._baudrate = DEFAULT_BAUDRATE
        if "port" in self._args:
            self._port = self._args.port
        else:
            self._port = DEFAULT_PORT
        if "server" in self._args:
            self._server = self._args.server
        else:
            self._server = None

        # are we a server or a client, or do we just not know?
        if CommBuffer.is_built:
            self._is_server = CommBuffer.is_server()
        else:
            self._is_server = False
        log.debug(self._args)

    def send(self) -> None:
        repeat = self._args.repeat if "repeat" in self._args else 1
        delay = self._args.delay if "delay" in self._args else 0
        duration = self._args.duration if "duration" in self._args else 0
        interval = self._args.interval if "interval" in self._args and self._args.interval else None
        self.command.send(
            repeat=repeat,
            delay=delay,
            duration=duration,
            interval=interval,
            baudrate=self._baudrate,
            port=self._port,
            server=self._server,
        )

    @property
    def command(self) -> CommandBase:
        return self._command

    @property
    def command_line(self) -> List[str]:
        return self._command_line

    @property
    def args(self) -> Namespace:
        return self._args

    @property
    def do_fire(self) -> bool:
        return self._do_fire

    @property
    def is_tmcc1(self) -> bool:
        return self._command_format == CommandSyntax.TMCC

    @property
    def is_tmcc2(self) -> bool:
        return self._command_format == CommandSyntax.LEGACY

    @property
    def command_format(self) -> CommandSyntax:
        return self._command_format

    @property
    def is_server(self) -> bool:
        return self._is_server

    @staticmethod
    def _validate_speed(arg: Any) -> int:
        try:
            return int(arg)  # try convert to int
        except ValueError:
            pass
        uc_arg = str(arg).upper()
        if uc_arg in TMCC2_SPEED_MAP:
            # feels a little hacky, but need a way to use a different map for TMCC commands
            if "-tmcc" in sys.argv or "-tmcc1" in sys.argv:
                return TMCC1_SPEED_MAP[uc_arg]
            return TMCC2_SPEED_MAP[uc_arg]
        raise ArgumentTypeError("Speed must be between 0 and 199 (0 and 31, for tmcc)")

    @staticmethod
    def _validate_time(arg: Any) -> float:
        try:
            arg = float(arg)  # try convert to float
            if arg >= 0.0:
                return arg
        except ValueError:
            pass
        raise ArgumentTypeError("Delay/Duration must be >= 0.0")

    @staticmethod
    def _validate_interval(arg: Any) -> int:
        try:
            arg = int(arg)  # try convert to float
            if arg >= DEFAULT_DURATION_INTERVAL_MSEC:
                return arg
        except ValueError:
            pass
        raise ArgumentTypeError(f"Interval must be >= {DEFAULT_DURATION_INTERVAL_MSEC}")

    @staticmethod
    def _validate_repeat(arg: Any) -> int:
        try:
            arg = int(arg)  # try convert to int
            if arg >= 1:
                return arg
        except ValueError:
            pass
        raise ArgumentTypeError("Repeat must be >= 1")


class CliBaseTMCC(CliBase):
    __metaclass__ = ABCMeta

    def __init__(self, arg_parser: ArgumentParser, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        if "format" in self._args and self._args.format:
            self._command_format = self._args.format
        else:
            self._command_format = CommandSyntax.LEGACY

    def _determine_scope(self):
        return CommandScope.TRAIN if self._args.train else CommandScope.ENGINE

    @property
    def is_train_command(self) -> bool:
        if "train" in self._args:
            return bool(self._args.train)
        else:
            return False


class DataAction(Action):
    """
    Custom action that sets both the command_def and data fields
    with the command_def value specified by 'const', and the data value
    specified by the user-provided argument or 'default'
    """

    def __init__(self, option_strings, dest, **kwargs):
        """
        We need to capture both the values of const and default, as we use the
        'const' value to specify the command_op to execute if this action is taken.
        Once saved, we reset its value to the default value.
        """
        self._default = kwargs.get("default")
        self._command_op = kwargs.get("const")
        kwargs["const"] = self._default
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, psr, namespace, values, option_string=None) -> None:
        setattr(namespace, self.dest, self._command_op)
        if values is not None:
            setattr(namespace, "data", values)
        else:
            setattr(namespace, "data", self._default)



from __future__ import annotations

import abc
from enum import unique
from typing import Dict, Tuple

from range_key_dict import RangeKeyDict

from ..command_def import CommandDef, CommandDefEnum
from ..constants import (
    DEFAULT_ADDRESS,
    DEFAULT_ENGINE_LABOR,
    RELATIVE_SPEED_MAP,
    CommandPrefix,
    CommandScope,
    CommandSyntax,
    OfficialRRSpeeds,
)

LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX: int = 0xFA
LEGACY_MULTIBYTE_COMMAND_PREFIX: int = 0xFB
LEGACY_ENGINE_COMMAND_PREFIX: int = 0xF8
LEGACY_TRAIN_COMMAND_PREFIX: int = 0xF9

"""
    TMCC2 constants
"""


class TMCC2Enum(CommandDefEnum):
    """
    Marker Interface for all TMCC2 enums
    """

    pass


"""
    Legacy/TMCC2 Protocol/first_byte Constants
"""


# All Legacy/TMCC2 commands begin with one of the following 1 byte sequences
# Engine/Train/Parameter 2 digit address is first 7 bits of first byte


@unique
class TMCC2CommandPrefix(CommandPrefix):
    ENGINE = LEGACY_ENGINE_COMMAND_PREFIX
    TRAIN = LEGACY_TRAIN_COMMAND_PREFIX
    ROUTE = LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX  # probably used for other things
    PARAMETER = LEGACY_MULTIBYTE_COMMAND_PREFIX


TMCC2_FIRST_BYTE_TO_SCOPE_MAP = {
    TMCC2CommandPrefix.ENGINE: CommandScope.ENGINE,
    TMCC2CommandPrefix.TRAIN: CommandScope.TRAIN,
    TMCC2CommandPrefix.ROUTE: CommandScope.ROUTE,
}

TMCC2_SCOPE_TO_FIRST_BYTE_MAP = {s: p for p, s in TMCC2_FIRST_BYTE_TO_SCOPE_MAP.items()}


class TMCC2CommandDef(CommandDef):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        command_bits: int,
        scope: CommandScope = CommandScope.ENGINE,
        is_addressable: bool = True,
        d_min: int = 0,
        d_max: int = 0,
        d_map: Dict[int, int] = None,
        alias: str = None,
        data: int = None,
        filtered=False,
        interval: int = None,
        d4_broadcast: bool = False,  # if True, command is broadcast from Base 3 for D4 engines
    ) -> None:
        super().__init__(
            command_bits,
            is_addressable,
            d_min=d_min,
            d_max=d_max,
            d_map=d_map,
            alias=alias,
            data=data,
            filtered=filtered,
            interval=interval,
        )
        self._scope = scope
        self._d4_broadcast = d4_broadcast

    @property
    def first_byte(self) -> bytes:
        return TMCC2_SCOPE_TO_FIRST_BYTE_MAP[self.scope].to_bytes(1, byteorder="big")

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @property
    def syntax(self) -> CommandSyntax:
        return CommandSyntax.LEGACY

    @property
    def address_mask(self) -> int:
        return 0xFFFF & ~((2**self.num_address_bits - 1) << 9)

    @property
    def is_d4_broadcast(self) -> bool:
        return self._d4_broadcast

    def address_from_bytes(self, byte_data: bytes) -> int:
        if self.is_addressable:
            value = int.from_bytes(byte_data[0:2], byteorder="big")
            address = (0xFFFF & (~self.address_mask & value)) >> 9
            if len(byte_data) <= 2 and 1 <= address <= 99:
                return address
            elif len(byte_data) <= 6 and address == 0:
                add_str = ""
                for i in range(2, 6):
                    add_str += chr(byte_data[i])
                return int(add_str)
            raise AttributeError(f"Cannot decode address from bytes: {byte_data.hex()}")
        else:
            return DEFAULT_ADDRESS

    @property
    def alias(self) -> TMCC2EngineCommandEnum | Tuple[TMCC2EngineCommandEnum, int] | None:
        if self._alias is not None:
            if isinstance(self._alias, str):
                alias = TMCC2EngineCommandEnum.by_name(self._alias, raise_exception=True)
            else:
                raise ValueError(f"Cannot classify command alias: {self._alias} ({type(self._alias)})")
            if self._data is None:
                return alias
            else:
                return alias, self._data
        return None


TMCC2_HALT_COMMAND: int = 0x01AB


@unique
class TMCC2HaltCommandEnum(TMCC2Enum):
    HALT = TMCC2CommandDef(TMCC2_HALT_COMMAND, CommandScope.ENGINE, alias="SYSTEM_HALT", data=99)


# The TMCC2 route command is an undocumented "extended block command" (0xFA)
LEGACY_ROUTE_COMMAND: int = 0x00FD


@unique
class TMCC2RouteCommandEnum(TMCC2Enum):
    FIRE = TMCC2CommandDef(LEGACY_ROUTE_COMMAND, scope=CommandScope.ROUTE)


TMCC2_AUX1_OFF_COMMAND: int = 0x0108
TMCC2_AUX1_ON_COMMAND: int = 0x010B
TMCC2_AUX1_OPTION_ONE_COMMAND: int = 0x0109  # Cab 1 Aux1 button
TMCC2_AUX1_OPTION_TWO_COMMAND: int = 0x010A
TMCC2_AUX2_OFF_COMMAND: int = 0x010C
TMCC2_AUX2_ON_COMMAND: int = 0x010F
TMCC2_AUX2_OPTION_ONE_COMMAND: int = 0x010D  # Cab 1 Aux2 button
TMCC2_AUX2_OPTION_TWO_COMMAND: int = 0x010E
TMCC2_AUX3_OPTION_ONE_COMMAND: int = 0x013B
TMCC2_BELL_OFF_COMMAND: int = 0x01F4
TMCC2_BELL_ONE_SHOT_DING_COMMAND: int = 0x01F0  # encode ding in last 2 bits (0 - 3)
TMCC2_BELL_ON_COMMAND: int = 0x01F5
TMCC2_BELL_SLIDER_POSITION_COMMAND: int = 0x01B0  # encode position in last 3 bits (2 - 5)
TMCC2_BLOW_HORN_ONE_COMMAND: int = 0x011C
TMCC2_BLOW_HORN_TWO_COMMAND: int = 0x011F
TMCC2_BOOST_SPEED_COMMAND: int = 0x0104
TMCC2_BRAKE_SPEED_COMMAND: int = 0x0107
TMCC2_DIESEL_RPM_SOUND_COMMAND: int = 0x01A0  # run level 0-7 encoded in the last 3 bits
TMCC2_ENGINE_LABOR_COMMAND: int = 0x01C0  # engine labor 0-31 encoded in the last 5 bits
TMCC2_ENG_AUGER_SOUND_COMMAND: int = 0x01F7
TMCC2_ENG_BRAKE_AIR_RELEASE_SOUND_COMMAND: int = 0x01F8
TMCC2_ENG_BRAKE_SQUEAL_SOUND_COMMAND: int = 0x01F6
TMCC2_ENG_CYLINDER_HISS_SOUND_COMMAND: int = 0x0152
TMCC2_ENG_LET_OFF_LONG_SOUND_COMMAND: int = 0x01FA
TMCC2_ENG_LET_OFF_SOUND_COMMAND: int = 0x01F9
TMCC2_ENG_POP_OFF_SOUND_COMMAND: int = 0x0153
TMCC2_ENG_REFUELLING_SOUND_COMMAND: int = 0x012D
TMCC2_FORWARD_DIRECTION_COMMAND: int = 0x0100
TMCC2_HIGHBALL_SPEED: int = 199
TMCC2_LIMITED_SPEED: int = 118
TMCC2_MEDIUM_SPEED: int = 92
TMCC2_MOTION_START_COMMAND: int = 0x00FA
TMCC2_MOTION_STOP_COMMAND: int = 0x00FE
TMCC2_NORMAL_SPEED: int = 145
TMCC2_NUMERIC_COMMAND: int = 0x0110
TMCC2_ENG_RPM_DOWN_COMMAND: int = 0x0116
TMCC2_ENG_RPM_UP_COMMAND: int = 0x0113
TMCC2_OPEN_FRONT_COUPLER_COMMAND: int = 0x0105
TMCC2_OPEN_REAR_COUPLER_COMMAND: int = 0x0106
TMCC2_QUILLING_HORN_COMMAND: int = 0x01E0
TMCC2_RESTRICTED_SPEED: int = 24
TMCC2_REVERSE_DIRECTION_COMMAND: int = 0x0103
TMCC2_RING_BELL_COMMAND: int = 0x011D
TMCC2_ROLL_SPEED: int = 1  # express speeds as simple integers
TMCC2_SET_ABSOLUTE_SPEED_COMMAND: int = 0x0000  # encode speed in last byte (0-199)
TMCC2_SET_ADDRESS_COMMAND: int = 0x012B
TMCC2_SET_BOOST_LEVEL_COMMAND: int = 0x00E8  # encode boost level in last 3 bits (0-7)
TMCC2_SET_BRAKE_LEVEL_COMMAND: int = 0x00E0  # encode brake level in last 3 bits (0-7)
TMCC2_SET_MOMENTUM_COMMAND: int = 0x00C8  # encode momentum in last 3 bits (0 - 7)
TMCC2_SET_MOMENTUM_LOW_COMMAND: int = 0x0128
TMCC2_SET_MOMENTUM_MEDIUM_COMMAND: int = 0x0129
TMCC2_SET_MOMENTUM_HIGH_COMMAND: int = 0x012A
TMCC2_SET_RELATIVE_SPEED_COMMAND: int = 0x0140  # Relative Speed -5 - 5 encoded in last 4 bits (offset by 5)
TMCC2_SET_TRAIN_BRAKE_COMMAND: int = 0x00F0  # encode train brake in the last 3 bits (0-7)
TMCC2_SHUTDOWN_SEQ_ONE_COMMAND: int = 0x01FD
TMCC2_SHUTDOWN_SEQ_TWO_COMMAND: int = 0x01FE
TMCC2_SLOW_SPEED: int = 59
TMCC2_SOUND_OFF_COMMAND: int = 0x0150
TMCC2_SOUND_ON_COMMAND: int = 0x0151
TMCC2_STALL_COMMAND: int = 0x00F8
TMCC2_START_UP_SEQ_ONE_COMMAND: int = 0x01FB
TMCC2_START_UP_SEQ_TWO_COMMAND: int = 0x01FC
TMCC2_STOP_IMMEDIATE_COMMAND: int = 0x00FB
TMCC2_TOGGLE_DIRECTION_COMMAND: int = 0x0101
TMCC2_VOLUME_DOWN_COMMAND_HACK = 0x0114
TMCC2_VOLUME_UP_COMMAND_HACK = 0x0111
TMCC2_WATER_INJECTOR_SOUND_COMMAND: int = 0x01A8

TMCC2_ASSIGN_SINGLE_FORWARD: int = 0x0120
TMCC2_ASSIGN_SINGLE_REVERSE: int = 0x0121
TMCC2_ASSIGN_HEAD_FORWARD: int = 0x0122
TMCC2_ASSIGN_HEAD_REVERSE: int = 0x0123
TMCC2_ASSIGN_MIDDLE_FORWARD: int = 0x0124
TMCC2_ASSIGN_MIDDLE_REVERSE: int = 0x0125
TMCC2_ASSIGN_REAR_FORWARD: int = 0x0126
TMCC2_ASSIGN_REAR_REVERSE: int = 0x0127
TMCC2_ASSIGN_CLEAR_CONSIST: int = 0x012C
TMCC2_ASSIGN_TO_TRAIN: int = 0x0130


TMCC2_SPEED_MAP = dict(
    STOP_HOLD=0,
    SH=0,
    STOP=0,
    ROLL=TMCC2_ROLL_SPEED,
    RO=TMCC2_ROLL_SPEED,
    RESTRICTED=TMCC2_RESTRICTED_SPEED,
    RE=TMCC2_RESTRICTED_SPEED,
    SLOW=TMCC2_SLOW_SPEED,
    SL=TMCC2_SLOW_SPEED,
    MEDIUM=TMCC2_MEDIUM_SPEED,
    ME=TMCC2_MEDIUM_SPEED,
    LIMITED=TMCC2_LIMITED_SPEED,
    LI=TMCC2_LIMITED_SPEED,
    NORMAL=TMCC2_NORMAL_SPEED,
    NO=TMCC2_NORMAL_SPEED,
    HIGH=TMCC2_HIGHBALL_SPEED,
    HIGHBALL=TMCC2_HIGHBALL_SPEED,
    HI=TMCC2_HIGHBALL_SPEED,
)

TMCC2_NUMERIC_SPEED_TO_DIRECTIVE_MAP = {s: p for p, s in TMCC2_SPEED_MAP.items() if len(p) > 2 and p != "HIGH"}


@unique
class TMCC2RRSpeedsEnum(OfficialRRSpeeds):
    STOP_HOLD = range(0, TMCC2_ROLL_SPEED)
    ROLL = range(TMCC2_ROLL_SPEED, TMCC2_RESTRICTED_SPEED)
    RESTRICTED = range(TMCC2_RESTRICTED_SPEED, TMCC2_SLOW_SPEED)
    SLOW = range(TMCC2_SLOW_SPEED, TMCC2_MEDIUM_SPEED)
    MEDIUM = range(TMCC2_MEDIUM_SPEED, TMCC2_LIMITED_SPEED)
    LIMITED = range(TMCC2_LIMITED_SPEED, TMCC2_NORMAL_SPEED)
    NORMAL = range(TMCC2_NORMAL_SPEED, TMCC2_HIGHBALL_SPEED)
    HIGHBALL = range(TMCC2_HIGHBALL_SPEED, TMCC2_HIGHBALL_SPEED + 1)


@unique
class TMCC2EngineCommandEnum(TMCC2Enum):
    ABSOLUTE_SPEED = TMCC2CommandDef(TMCC2_SET_ABSOLUTE_SPEED_COMMAND, d_max=199, filtered=True, d4_broadcast=True)
    AUGER = TMCC2CommandDef(TMCC2_ENG_AUGER_SOUND_COMMAND)
    AUX1_OFF = TMCC2CommandDef(TMCC2_AUX1_OFF_COMMAND)
    AUX1_ON = TMCC2CommandDef(TMCC2_AUX1_ON_COMMAND)
    AUX1_OPTION_ONE = TMCC2CommandDef(TMCC2_AUX1_OPTION_ONE_COMMAND)
    AUX1_OPTION_TWO = TMCC2CommandDef(TMCC2_AUX1_OPTION_TWO_COMMAND)
    AUX2_OFF = TMCC2CommandDef(TMCC2_AUX2_OFF_COMMAND)
    AUX2_ON = TMCC2CommandDef(TMCC2_AUX2_ON_COMMAND)
    AUX2_OPTION_ONE = TMCC2CommandDef(TMCC2_AUX2_OPTION_ONE_COMMAND)
    AUX2_OPTION_TWO = TMCC2CommandDef(TMCC2_AUX2_OPTION_TWO_COMMAND)
    AUX3_OPTION_ONE = TMCC2CommandDef(TMCC2_AUX3_OPTION_ONE_COMMAND)
    BELL_OFF = TMCC2CommandDef(TMCC2_BELL_OFF_COMMAND)
    BELL_ON = TMCC2CommandDef(TMCC2_BELL_ON_COMMAND)
    BELL_ONE_SHOT_DING = TMCC2CommandDef(TMCC2_BELL_ONE_SHOT_DING_COMMAND, d_max=3, interval=1000)
    BELL_SLIDER_POSITION = TMCC2CommandDef(TMCC2_BELL_SLIDER_POSITION_COMMAND, d_min=2, d_max=5)
    BLOW_HORN_ONE = TMCC2CommandDef(TMCC2_BLOW_HORN_ONE_COMMAND, interval=100)
    BLOW_HORN_TWO = TMCC2CommandDef(TMCC2_BLOW_HORN_TWO_COMMAND, interval=100)
    BOOST_LEVEL = TMCC2CommandDef(TMCC2_SET_BOOST_LEVEL_COMMAND, d_max=7)
    BOOST_SPEED = TMCC2CommandDef(TMCC2_BOOST_SPEED_COMMAND, interval=200)
    BRAKE_AIR_RELEASE = TMCC2CommandDef(TMCC2_ENG_BRAKE_AIR_RELEASE_SOUND_COMMAND)
    BRAKE_LEVEL = TMCC2CommandDef(TMCC2_SET_BRAKE_LEVEL_COMMAND, d_max=7)
    BRAKE_SPEED = TMCC2CommandDef(TMCC2_BRAKE_SPEED_COMMAND, interval=200)
    BRAKE_SQUEAL = TMCC2CommandDef(TMCC2_ENG_BRAKE_SQUEAL_SOUND_COMMAND)
    CYLINDER_HISS = TMCC2CommandDef(TMCC2_ENG_CYLINDER_HISS_SOUND_COMMAND)
    DIESEL_RPM = TMCC2CommandDef(TMCC2_DIESEL_RPM_SOUND_COMMAND, d_max=7, filtered=True)
    ENGINE_LABOR = TMCC2CommandDef(TMCC2_ENGINE_LABOR_COMMAND, d_max=31, filtered=True, d4_broadcast=True)
    ENGINE_LABOR_DEFAULT = TMCC2CommandDef(
        TMCC2_ENGINE_LABOR_COMMAND | 12, alias="ENGINE_LABOR", data=DEFAULT_ENGINE_LABOR, filtered=True
    )
    FORWARD_DIRECTION = TMCC2CommandDef(TMCC2_FORWARD_DIRECTION_COMMAND, filtered=True, d4_broadcast=True)
    FRONT_COUPLER = TMCC2CommandDef(TMCC2_OPEN_FRONT_COUPLER_COMMAND)
    LET_OFF = TMCC2CommandDef(TMCC2_ENG_LET_OFF_SOUND_COMMAND)
    LET_OFF_LONG = TMCC2CommandDef(TMCC2_ENG_LET_OFF_LONG_SOUND_COMMAND)
    MOMENTUM = TMCC2CommandDef(TMCC2_SET_MOMENTUM_COMMAND, d_max=7)
    MOMENTUM_HIGH = TMCC2CommandDef(TMCC2_SET_MOMENTUM_HIGH_COMMAND)
    MOMENTUM_LOW = TMCC2CommandDef(TMCC2_SET_MOMENTUM_LOW_COMMAND)
    MOMENTUM_MEDIUM = TMCC2CommandDef(TMCC2_SET_MOMENTUM_MEDIUM_COMMAND)
    MOTION_START = TMCC2CommandDef(TMCC2_MOTION_START_COMMAND)
    MOTION_STOP = TMCC2CommandDef(TMCC2_MOTION_STOP_COMMAND)
    NUMERIC = TMCC2CommandDef(TMCC2_NUMERIC_COMMAND, d_max=9)
    POP_OFF = TMCC2CommandDef(TMCC2_ENG_POP_OFF_SOUND_COMMAND)
    QUILLING_HORN = TMCC2CommandDef(TMCC2_QUILLING_HORN_COMMAND, d_max=15, interval=100)
    REAR_COUPLER = TMCC2CommandDef(TMCC2_OPEN_REAR_COUPLER_COMMAND)
    REFUELLING = TMCC2CommandDef(TMCC2_ENG_REFUELLING_SOUND_COMMAND)
    RELATIVE_SPEED = TMCC2CommandDef(TMCC2_SET_RELATIVE_SPEED_COMMAND, d_map=RELATIVE_SPEED_MAP)
    REVERSE_DIRECTION = TMCC2CommandDef(TMCC2_REVERSE_DIRECTION_COMMAND, filtered=True, d4_broadcast=True)
    RING_BELL = TMCC2CommandDef(TMCC2_RING_BELL_COMMAND)
    RPM_DOWN = TMCC2CommandDef(TMCC2_ENG_RPM_DOWN_COMMAND, alias="NUMERIC", data=6)
    RPM_UP = TMCC2CommandDef(TMCC2_ENG_RPM_UP_COMMAND, alias="NUMERIC", data=3)
    SET_ADDRESS = TMCC2CommandDef(TMCC2_SET_ADDRESS_COMMAND)
    SHUTDOWN_DELAYED = TMCC2CommandDef(TMCC2_NUMERIC_COMMAND | 5, alias="NUMERIC", data=5)
    RESET = TMCC2CommandDef(TMCC2_NUMERIC_COMMAND | 0, alias="NUMERIC", data=0)
    SHUTDOWN_DELAYED_NOP = TMCC2CommandDef(TMCC2_SHUTDOWN_SEQ_ONE_COMMAND)
    SHUTDOWN_IMMEDIATE = TMCC2CommandDef(TMCC2_SHUTDOWN_SEQ_TWO_COMMAND)
    SOUND_OFF = TMCC2CommandDef(TMCC2_SOUND_OFF_COMMAND)
    SOUND_ON = TMCC2CommandDef(TMCC2_SOUND_ON_COMMAND)
    SPEED_HIGHBALL = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_HIGHBALL_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_HIGHBALL_SPEED
    )
    SPEED_LIMITED = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_LIMITED_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_LIMITED_SPEED
    )
    SPEED_MEDIUM = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_MEDIUM_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_MEDIUM_SPEED
    )
    SPEED_NORMAL = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_NORMAL_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_NORMAL_SPEED
    )
    SPEED_RESTRICTED = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_RESTRICTED_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_RESTRICTED_SPEED
    )
    SPEED_ROLL = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_ROLL_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_ROLL_SPEED
    )
    SPEED_SLOW = TMCC2CommandDef(
        TMCC2_SET_ABSOLUTE_SPEED_COMMAND | TMCC2_SLOW_SPEED, alias="ABSOLUTE_SPEED", data=TMCC2_SLOW_SPEED
    )
    SPEED_STOP_HOLD = TMCC2CommandDef(TMCC2_SET_ABSOLUTE_SPEED_COMMAND, alias="ABSOLUTE_SPEED", data=0)
    STALL = TMCC2CommandDef(TMCC2_STALL_COMMAND)
    START_UP_DELAYED = TMCC2CommandDef(TMCC2_START_UP_SEQ_ONE_COMMAND)
    START_UP_IMMEDIATE = TMCC2CommandDef(TMCC2_START_UP_SEQ_TWO_COMMAND)
    STOP_IMMEDIATE = TMCC2CommandDef(TMCC2_STOP_IMMEDIATE_COMMAND, filtered=True, d4_broadcast=True)
    SYSTEM_HALT = TMCC2CommandDef(TMCC2_HALT_COMMAND)
    TOGGLE_DIRECTION = TMCC2CommandDef(TMCC2_TOGGLE_DIRECTION_COMMAND, filtered=True, d4_broadcast=True)
    TOWER_CHATTER = TMCC2CommandDef(TMCC2_NUMERIC_COMMAND | 7, alias="NUMERIC", data=7)
    TRAIN_BRAKE = TMCC2CommandDef(TMCC2_SET_TRAIN_BRAKE_COMMAND, d_max=7, filtered=True, d4_broadcast=True)
    WATER_INJECTOR = TMCC2CommandDef(TMCC2_WATER_INJECTOR_SOUND_COMMAND)
    VOLUME_UP = TMCC2CommandDef(TMCC2_VOLUME_UP_COMMAND_HACK, alias="NUMERIC", data=1)
    VOLUME_DOWN = TMCC2CommandDef(TMCC2_VOLUME_DOWN_COMMAND_HACK, alias="NUMERIC", data=4)
    SINGLE_FORWARD = TMCC2CommandDef(TMCC2_ASSIGN_SINGLE_FORWARD)
    SINGLE_REVERSE = TMCC2CommandDef(TMCC2_ASSIGN_SINGLE_REVERSE)
    HEAD_FORWARD = TMCC2CommandDef(TMCC2_ASSIGN_HEAD_FORWARD)
    HEAD_REVERSE = TMCC2CommandDef(TMCC2_ASSIGN_HEAD_REVERSE)
    MIDDLE_FORWARD = TMCC2CommandDef(TMCC2_ASSIGN_MIDDLE_FORWARD)
    MIDDLE_REVERSE = TMCC2CommandDef(TMCC2_ASSIGN_MIDDLE_REVERSE)
    REAR_FORWARD = TMCC2CommandDef(TMCC2_ASSIGN_REAR_FORWARD)
    REAR_REVERSE = TMCC2CommandDef(TMCC2_ASSIGN_REAR_REVERSE)
    CLEAR_CONSIST = TMCC2CommandDef(TMCC2_ASSIGN_CLEAR_CONSIST)


# map dereferenced commands to their aliases
TMCC2_COMMAND_TO_ALIAS_MAP = {}
for tmcc2_enum in [TMCC2EngineCommandEnum, TMCC2HaltCommandEnum, TMCC2RouteCommandEnum]:
    for enum in tmcc2_enum:
        if enum.is_alias:
            TMCC2_COMMAND_TO_ALIAS_MAP[enum.alias] = enum

# Speed to RPM Map
TMCC2_SPEED_TO_RPM = RangeKeyDict(
    {
        (0, 4): 0,
        (4, 29): 1,
        (29, 57): 2,
        (57, 86): 3,
        (86, 114): 4,
        (114, 143): 5,
        (143, 171): 6,
        (171, 200): 7,
    }
)


def tmcc2_speed_to_rpm(speed: int) -> int:
    return TMCC2_SPEED_TO_RPM[speed]



from __future__ import annotations

import abc
import math
import sys
from abc import ABC
from enum import Enum
from typing import Any, Dict, Tuple, TypeVar

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from .constants import CommandScope, CommandSyntax, Mixins

E = TypeVar("E", bound="CommandDefEnum")


class CommandDef(ABC):
    __metaclass__ = abc.ABCMeta
    """
        Marker class for TMCC1 and TMCC2 Command Defs, allowing the CLI layer
        to work with them in a command format agnostic manner.
    """

    def __init__(
        self,
        command_bits: int,
        is_addressable: bool = True,
        num_address_bits: int = 7,
        d_min: int = 0,
        d_max: int = 0,
        d_map: Dict[int, int] = None,
        do_reverse_lookup: bool = True,
        alias: str = None,
        data: int = None,
        filtered: bool = False,
        interval: int = None,
    ) -> None:
        self._command_bits: int = command_bits
        self._is_addressable = is_addressable
        self._num_address_bits = num_address_bits
        self._do_reverse_lookup = do_reverse_lookup
        self._alias: str = alias.strip().upper() if alias else None
        self._data: int = data
        self._d_min = d_min
        self._d_max = d_max
        self._d_map = d_map
        self._d_bits = 0
        if d_max:
            self._d_bits = math.ceil(math.log2(d_max))
        elif d_map is not None:
            self._d_bits = math.ceil(math.log2(max(d_map.values())))
        self._filtered = filtered
        self._interval = interval

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} 0x{self.bits:04x}: {self.num_data_bits} data bits"

    @property
    def bits(self) -> int:
        return self._command_bits

    @property
    def as_bytes(self) -> bytes:
        return self.bits.to_bytes(3, byteorder="big")

    @property
    def is_addressable(self) -> bool:
        return self._is_addressable

    @property
    def is_data(self) -> bool:
        return self.num_data_bits != 0

    @property
    def is_filtered(self) -> bool:
        return self._filtered

    @property
    def interval(self) -> int:
        return self._interval

    @property
    def is_aux1_prefixed(self) -> bool:
        return False

    def is_valid_data(self, candidate: int, from_bytes: bool = False) -> bool:
        """
        Determine if a candidate value is valid, given the constraints on this
        CommandDef. If from_bytes is True, the candidate value will be treated
        as coming from a bytes value, and the values (not keys) in data_map are
        used to validate.
        """
        if self.is_data is True:
            if self.data_map:
                if from_bytes is True:
                    return candidate in [v for k, v in self.data_map.items()]
                elif from_bytes is False:
                    return candidate in self.data_map
            else:
                return self.data_min <= candidate <= self.data_max
        return False

    def data_from_bytes(self, byte_data: bytes) -> int:
        if self.is_data:
            value = int.from_bytes(byte_data, byteorder="big")
            data = 0xFFFF & (~self.data_mask & value)
            if self.data_map:
                for k, v in self.data_map.items():
                    if v == data:
                        return k
            else:
                return data
        return 0

    @property
    def is_alias(self) -> bool:
        return self._alias is not None

    @property
    def num_data_bits(self) -> int:
        return self._d_bits

    @property
    def num_address_bits(self) -> int:
        if self.is_addressable:
            return self._num_address_bits
        else:
            return 0

    @property
    def data_mask(self) -> int:
        return 0xFFFF & ~(2**self.num_data_bits - 1)

    @property
    def data_min(self) -> int:
        return self._d_min

    @property
    def data_max(self) -> int:
        return self._d_max

    @property
    def data_map(self) -> Dict[int, int] | None:
        return self._d_map

    @property
    def syntax(self) -> CommandSyntax:
        raise TypeError(f"Invalid command syntax: {self}")

    @property
    def is_tmcc1(self) -> bool:
        return self.syntax == CommandSyntax.TMCC

    @property
    def is_tmcc2(self) -> bool:
        return self.syntax == CommandSyntax.LEGACY

    @property
    def is_legacy(self) -> bool:
        return self.is_tmcc2

    @property
    def identifier(self) -> int | None:
        """
        Only relevant for TMCC1-style commands
        """
        return None

    @property
    def do_reverse_lookup(self) -> bool:
        return self._do_reverse_lookup

    @property
    @abc.abstractmethod
    def address_mask(self) -> int | None:
        return None

    @property
    @abc.abstractmethod
    def first_byte(self) -> bytes | None:
        return None

    @property
    @abc.abstractmethod
    def scope(self) -> CommandScope | None:
        return None

    @property
    @abc.abstractmethod
    def alias(self) -> E | Tuple[E, int] | None:
        return None


class CommandDefMixins(Mixins):
    @classmethod
    def by_value(cls, value: Any, raise_exception: bool = False) -> Self | None:
        """
        We redefine by_value to allow handling of command defs
        """
        for _, member in cls.__members__.items():
            if member.value == value:
                return member
            if hasattr(member.value, "bits") and member.value.bits == value:
                return member
            if isinstance(member, CommandDef) and CommandDef(value).bits == value:
                return member
            if isinstance(member, CommandDefEnum) and isinstance(value, int) and not member.value.is_alias:
                cd = member.value  # CommandDef
                if cd.is_data is True:
                    data = 0xFFFF & (~cd.data_mask & value)
                    if value & cd.address_mask & cd.data_mask == member.value.bits and cd.is_valid_data(
                        data, from_bytes=True
                    ):
                        return member
                elif value & cd.address_mask == member.value.bits:
                    return member
        if raise_exception:
            raise ValueError(f"'{value}' is not a valid {cls.__name__}")
        else:
            return None


class CommandDefEnum(CommandDefMixins, Enum):
    """
    Marker Interface to allow TMCC1EngineCommandDef and TMCC2EngineCommandDef enums
    to be handled by engine commands
    """

    @property
    def command_def(self) -> CommandDef:
        return self.value

    @property
    def scope(self) -> CommandScope:
        return self.command_def.scope

    @property
    def syntax(self) -> CommandSyntax:
        return self.command_def.syntax

    @property
    def is_tmcc1(self) -> bool:
        return self.command_def.is_tmcc1

    @property
    def is_tmcc2(self) -> bool:
        return self.command_def.is_tmcc2

    @property
    def is_legacy(self) -> bool:
        return self.command_def.is_tmcc2

    @property
    def bits(self) -> int:
        return self.value.bits

    @property
    def is_alias(self) -> bool:
        return self.command_def.is_alias

    @property
    def is_filtered(self) -> bool:
        return self.command_def.is_filtered

    @property
    def alias(self) -> E | Tuple[E, int]:
        return self.command_def.alias

    @property
    def as_bytes(self) -> bytes:
        return self.value.as_bytes



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..db.component_state_store import ComponentStateStore
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from .gpio_device import GpioDevice, P
from .state_source import AccessoryStateSource


class PowerDistrict(GpioDevice):
    def __init__(
        self,
        address: int,
        on_pin: P,
        off_pin: P,
        on_led_pin: P = None,
        cathode: bool = True,
        initial_state: TMCC1AuxCommandEnum | bool = None,
    ) -> None:
        """
        Control a power district that responds to TMCC1 accessory commands, such
        as an LCS BP2 configured in "Acc" mode.
        """
        if initial_state is None:
            state = ComponentStateStore.get_state(CommandScope.ACC, address, create=False)
            if state:
                initial_state = state.aux_state
            if initial_state is None:
                # last resort, assume district is off
                initial_state = TMCC1AuxCommandEnum.AUX2_OPT_ONE

        # make the CommandReqs
        on_req, self.on_btn, self.on_led = self.make_button(
            on_pin,
            TMCC1AuxCommandEnum.AUX1_OPT_ONE,
            address,
            led_pin=on_led_pin,
            cathode=cathode,
            initially_on=initial_state == TMCC1AuxCommandEnum.AUX1_OPT_ONE,
        )
        off_req, self.off_btn, self.off_led = self.make_button(
            off_pin,
            TMCC1AuxCommandEnum.AUX2_OPT_ONE,
            address,
            cathode=cathode,
            initially_on=initial_state == TMCC1AuxCommandEnum.AUX2_OPT_ONE,
        )
        # bind actions to buttons
        on_action = on_req.as_action(repeat=2)
        off_action = off_req.as_action(repeat=2)

        self.on_btn.when_pressed = self.with_on_action(on_action, self.on_led)
        self.off_btn.when_pressed = self.with_off_action(off_action, self.on_led)

        if self.on_led:
            # listen for external state changes
            self.cache_handler(AccessoryStateSource(address, self.on_led, aux_state=TMCC1AuxCommandEnum.AUX1_OPT_ONE))



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from argparse import ArgumentParser
from typing import List

from ..protocol.multibyte.dialog_cmd import DialogCmd
from ..protocol.multibyte.multibyte_constants import TMCC2RailSoundsDialogControl
from ..utils.argument_parser import PyTrainArgumentParser
from . import CliBaseTMCC

log = logging.getLogger(__name__)


class DialogsCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls):
        dialog_parser = PyTrainArgumentParser(add_help=False)
        dialog_parser.add_argument("engine", metavar="Engine/Train/Car", type=int, help="Engine/Train/Car to control")
        dialog_parser.add_argument(
            "-emergency",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.EMERGENCY_CONTEXT_DEPENDENT,
            help="Emergency context-dependent dialog",
        )
        dialog_parser.add_argument(
            "-sd",
            "-conventional_shutdown",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.CONVENTIONAL_SHUTDOWN,
            help="Conventional mode SHUTDOWN dialog",
        )
        dialog_parser.add_argument(
            "-sh",
            "-short_horn",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.SHORT_HORN,
            help="Conventional mode short horn dialog",
        )
        dialog_parser.add_argument(
            "-2",
            "-scene_two",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.SCENE_TWO,
            help="Scene <2> key context-dependent trigger",
        )
        dialog_parser.add_argument(
            "-5",
            "-scene_five",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.SCENE_FIVE,
            help="Scene <5> key context-dependent trigger",
        )
        dialog_parser.add_argument(
            "-7",
            "-scene_seven",
            action="store_const",
            dest="option",
            const=TMCC2RailSoundsDialogControl.SCENE_SEVEN,
            help="Scene <7> key context-dependent trigger",
        )
        sp = dialog_parser.add_subparsers(dest="sub_command", help="Engine/train sub-commands")

        eng = sp.add_parser("engineer", aliases=["en"], help="Engineer dialogs")
        eng_group = eng.add_mutually_exclusive_group()
        eng_group.add_argument(
            "-ac",
            "-all_clear",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ALL_CLEAR,
            dest="option",
            help="Engineer: first all clear",
        )
        eng_group.add_argument(
            "-ar",
            "-arriving",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ARRIVING,
            dest="option",
            help="Engineer: first departure denied",
        )
        eng_group.add_argument(
            "-ad",
            "-arrived",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ARRIVED,
            dest="option",
            help="Engineer: arrived",
        )
        eng_group.add_argument(
            "-ack",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK,
            dest="option",
            help="Engineer: acknowledge communication",
        )
        eng_group.add_argument(
            "-asb",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_STAND_BY,
            dest="option",
            help="Engineer: ACK standing by",
        )
        eng_group.add_argument(
            "-aac",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_CLEARED,
            dest="option",
            help="Engineer: ACK cleared to go",
        )
        eng_group.add_argument(
            "-aca",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_CLEAR_AHEAD,
            dest="option",
            help="Engineer: ACK clear ahead",
        )
        eng_group.add_argument(
            "-aci",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_CLEAR_INBOUND,
            dest="option",
            help="Engineer: ACK clear inbound",
        )
        eng_group.add_argument(
            "-awb",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_WELCOME_BACK,
            dest="option",
            help="Engineer: ACK welcome back",
        )
        eng_group.add_argument(
            "-aid",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ACK_ID,
            dest="option",
            help="Engineer: ACK identify & out",
        )
        eng_group.add_argument(
            "-dd",
            "-departure_denied",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_DEPARTURE_DENIED,
            dest="option",
            help="Engineer: first departure denied",
        )
        eng_group.add_argument(
            "-dg",
            "-departure_granted",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_DEPARTURE_GRANTED,
            dest="option",
            help="Engineer: first departure granted",
        )
        eng_group.add_argument(
            "-de",
            "-departed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_DEPARTED,
            dest="option",
            help="Engineer: first have departed",
        )
        eng_group.add_argument(
            "-fl",
            "-fuel_level",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_FUEL_LEVEL,
            dest="option",
            help="Engineer: speaks fuel level",
        )
        eng_group.add_argument(
            "-fr",
            "-fuel_refilled",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_FUEL_REFILLED,
            dest="option",
            help="Engineer: speaks fuel refilled",
        )
        eng_group.add_argument(
            "-id",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_ID,
            dest="option",
            help="Engineer: identify",
        )
        eng_group.add_argument(
            "-s",
            "-speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED,
            dest="option",
            help="Engineer: speaks speed",
        )
        eng_group.add_argument(
            "-sd",
            "-shutdown",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SHUTDOWN,
            dest="option",
            help="Engineer: shut down",
        )
        eng_group.add_argument(
            "-sh",
            "-stop_hold",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_STOP_HOLD,
            dest="option",
            help="Engineer: first stop and hold ack",
        )
        eng_group.add_argument(
            "-sr",
            "-restricted_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_RESTRICTED,
            dest="option",
            help="Engineer: first restricted speed ack",
        )
        eng_group.add_argument(
            "-ss",
            "-std_step_to_data",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_SLOW,
            dest="option",
            help="Engineer: first slow speed ack",
        )
        eng_group.add_argument(
            "-sm",
            "-medium_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_MEDIUM,
            dest="option",
            help="Engineer: first medium speed ack",
        )
        eng_group.add_argument(
            "-sl",
            "-limited_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_LIMITED,
            dest="option",
            help="Engineer: first limited speed ack",
        )
        eng_group.add_argument(
            "-sn",
            "-normal_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_NORMAL,
            dest="option",
            help="Engineer: first normal speed ack",
        )
        eng_group.add_argument(
            "-sx",
            "-highball_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_SPEED_HIGHBALL,
            dest="option",
            help="Engineer: first highball speed ack",
        )
        eng_group.add_argument(
            "-wl",
            "-water_level",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_WATER_LEVEL,
            dest="option",
            help="Engineer: speaks water level",
        )
        eng_group.add_argument(
            "-wr",
            "-water_refilled",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.ENGINEER_WATER_REFILLED,
            dest="option",
            help="Engineer: speaks water refilled",
        )

        tower = sp.add_parser("tower", aliases=["to"], help="Tower dialogs")
        tower_group = tower.add_mutually_exclusive_group()
        tower_group.add_argument(
            "-su",
            "-start_up",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_STARTUP,
            dest="option",
            help="Tower: first engine startup dialog",
        )
        tower_group.add_argument(
            "-sh",
            "-stop_hold",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_STOP_HOLD,
            dest="option",
            help="Tower: first stop and hold dialog (non-emergency)",
        )
        tower_group.add_argument(
            "-sr",
            "-restricted_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_RESTRICTED,
            dest="option",
            help="Tower: first restricted speed",
        )
        tower_group.add_argument(
            "-ss",
            "-std_step_to_data",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_SLOW,
            dest="option",
            help="Tower: first slow speed",
        )
        tower_group.add_argument(
            "-sm",
            "-medium_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_MEDIUM,
            dest="option",
            help="Tower: first medium speed",
        )
        tower_group.add_argument(
            "-sl",
            "-limited_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_LIMITED,
            dest="option",
            help="Tower: first limited speed",
        )
        tower_group.add_argument(
            "-sn",
            "-normal_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_NORMAL,
            dest="option",
            help="Tower: first normal speed",
        )
        tower_group.add_argument(
            "-sx",
            "-highball_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SPEED_HIGHBALL,
            dest="option",
            help="Tower: first highball speed",
        )
        tower_group.add_argument(
            "-dd",
            "-departure_denied",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_DEPARTURE_DENIED,
            dest="option",
            help="Tower: departure denied",
        )
        tower_group.add_argument(
            "-dg",
            "-departure_granted",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_DEPARTURE_GRANTED,
            dest="option",
            help="Tower: departure granted",
        )
        tower_group.add_argument(
            "-de",
            "-departed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_DEPARTED,
            dest="option",
            help="Tower: departed",
        )
        tower_group.add_argument(
            "-ac",
            "-all_clear",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_ALL_CLEAR,
            dest="option",
            help="Tower: all clear",
        )
        tower_group.add_argument(
            "-ar",
            "-arriving",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_ARRIVING,
            dest="option",
            help="Tower: arriving",
        )
        tower_group.add_argument(
            "-ad",
            "-arrived",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_ARRIVED,
            dest="option",
            help="Tower: arrived",
        )
        tower_group.add_argument(
            "-sd",
            "-shutdown",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.TOWER_SHUTDOWN,
            dest="option",
            help="Tower: shut down",
        )
        seq = sp.add_parser("sequence", aliases=["sq"], help="Sequence control dialogs")
        seq_group = seq.add_mutually_exclusive_group()
        seq_group.add_argument(
            "-off",
            "-sequence_off",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_OFF,
            dest="option",
            help="SeqCtl: sequence control off",
        )
        seq_group.add_argument(
            "-on",
            "-sequence_on",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_ON,
            dest="option",
            help="SeqCtl: sequence control on",
        )
        seq_group.add_argument(
            "-c",
            "-clear",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_CLEAR,
            dest="option",
            help="SeqCtl: cleared out bound",
        )
        seq_group.add_argument(
            "-d",
            "-departed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_DEPARTED,
            dest="option",
            help="SeqCtl: have departed",
        )
        seq_group.add_argument(
            "-t",
            "-in_transit",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_TRANSIT,
            dest="option",
            help="SeqCtl: in transit",
        )
        seq_group.add_argument(
            "-x",
            "-max_speed",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.SEQUENCE_MAX_SPEED,
            dest="option",
            help="SeqCtl: max authorized speed",
        )

        cond = sp.add_parser("conductor", aliases=["co"], help="Conductor dialogs")
        cond_group = cond.add_mutually_exclusive_group()
        cond_group.add_argument(
            "-aa",
            "-all_aboard",
            action="store_const",
            const=TMCC2RailSoundsDialogControl.CONDUCTOR_ALL_ABOARD,
            dest="option",
            help="Conductor: all aboard",
        )

        return PyTrainArgumentParser(
            "Engine/train/car dialogs",
            parents=[dialog_parser, cls.multi_parser(), cls.train_parser(), cls.cli_parser()],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        engine: int = self._args.engine
        option = self._args.option
        try:
            scope = self._determine_scope()
            cmd = DialogCmd(
                engine,
                TMCC2RailSoundsDialogControl(option),
                0,
                scope,
                baudrate=self._baudrate,
                port=self._port,
                server=self._server,
            )
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



# noinspection PyPackageRequirements
import pytest

from src.pytrain.protocol.command_base import CommandBase


class TestCommandBase:
    def test_can_not_instantiate_command_base(self) -> None:
        with pytest.raises(TypeError):
            # noinspection PyTypeChecker
            CommandBase(None, None)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from .i2c.oled import OledDevice
from .launch_status import LaunchStatus
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum
from .gpio_device import GpioDevice, P


class LaunchPad(GpioDevice):
    def __init__(
        self,
        address: int = 39,
        gantry_fwd_pin: P = None,
        gantry_bck_pin: P = None,
        launch_seq_pin: P = None,
        launch_now_pin: P = None,
        launch_t15_pin: P = None,
        abort_pin: P = None,
        siren_pin: P = None,
        klaxon_pin: P = None,
        ground_crew_pin: P = None,
        mission_control_pin: P = None,
        flicker_on_pin: P = None,
        flicker_off_pin: P = None,
        title: str | None = "Launch Pad 39A",
        device: OledDevice | str = None,
        device_address: int = 0x3C,
        repeat_every: float = 0.02,
    ):
        super().__init__()
        # if a device is specified, set up oled display
        if device:
            if title is None and address == 39:
                title = "Pad 39A"
            self._lsd = LaunchStatus(address, title, device_address, device)
        else:
            self._lsd = None
        # use momentary contact switch to move the gantry
        left_cmd, self.gantry_fwd_btn, _ = self.make_button(
            gantry_fwd_pin,
            command=TMCC1EngineCommandEnum.RELATIVE_SPEED,
            address=address,
            data=-1,
            scope=CommandScope.ENGINE,
            hold_repeat=True,
            hold_time=repeat_every,
        )
        self.gantry_fwd_btn.when_pressed = left_cmd.as_action()
        self.gantry_fwd_btn.when_held = left_cmd.as_action()

        right_cmd, self.gantry_bck_btn, _ = self.make_button(
            gantry_bck_pin,
            command=TMCC1EngineCommandEnum.RELATIVE_SPEED,
            address=address,
            data=1,
            scope=CommandScope.ENGINE,
            hold_repeat=True,
            hold_time=repeat_every,
        )
        self.gantry_bck_btn.when_pressed = right_cmd.as_action()
        self.gantry_bck_btn.when_held = right_cmd.as_action()

        # launch sequence
        if launch_seq_pin:
            cmd, self.launch_seq_btn, _ = self.make_button(
                launch_seq_pin,
                command=TMCC1EngineCommandEnum.AUX1_OPTION_ONE,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.launch_seq_btn.when_pressed = cmd.as_action(duration=4.0)
        else:
            self.launch_seq_btn = None

        # launch sequence
        if launch_now_pin:
            cmd, self.launch_now_btn, _ = self.make_button(
                launch_now_pin,
                command=TMCC1EngineCommandEnum.FRONT_COUPLER,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.launch_now_btn.when_pressed = cmd.as_action()
        else:
            self.launch_now_btn = None

        # launch T-15
        if launch_t15_pin:
            cmd, self.launch_t15_btn, _ = self.make_button(
                launch_t15_pin,
                command=TMCC1EngineCommandEnum.REAR_COUPLER,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.launch_t15_btn.when_pressed = cmd.as_action()
        else:
            self.launch_t15_btn = None

        # launch abort
        if abort_pin:
            cmd, self.abort_btn, _ = self.make_button(
                abort_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=0,
                scope=CommandScope.ENGINE,
            )
            self.abort_btn.when_pressed = cmd.as_action()
        else:
            self.abort_btn = None

        # Siren
        if siren_pin:
            cmd, self.siren_btn, _ = self.make_button(
                siren_pin,
                command=TMCC1EngineCommandEnum.BLOW_HORN_ONE,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.siren_btn.when_pressed = cmd.as_action()
        else:
            self.siren_btn = None

        # Klaxon
        if klaxon_pin:
            cmd, self.klaxon_btn, _ = self.make_button(
                klaxon_pin,
                command=TMCC1EngineCommandEnum.RING_BELL,
                address=address,
                scope=CommandScope.ENGINE,
            )
            self.klaxon_btn.when_pressed = cmd.as_action()
        else:
            self.klaxon_btn = None

        # ground crew dialog
        if ground_crew_pin:
            cmd, self.ground_crew_btn, _ = self.make_button(
                ground_crew_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=2,
                scope=CommandScope.ENGINE,
            )
            self.ground_crew_btn.when_pressed = cmd.as_action()
        else:
            self.ground_crew_btn = None

        # mission control dialog
        if mission_control_pin:
            cmd, self.mission_control_btn, _ = self.make_button(
                mission_control_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=7,
                scope=CommandScope.ENGINE,
            )
            self.mission_control_btn.when_pressed = cmd.as_action()
        else:
            self.mission_control_btn = None

        # smoke/flicker on
        if flicker_on_pin:
            cmd, self.flicker_on_btn, _ = self.make_button(
                flicker_on_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=9,
                scope=CommandScope.ENGINE,
            )
            self.flicker_on_btn.when_pressed = cmd.as_action()
        else:
            self.flicker_on_btn = None

        # smoke/flicker off
        if flicker_off_pin:
            cmd, self.flicker_off_btn, _ = self.make_button(
                flicker_off_pin,
                command=TMCC1EngineCommandEnum.NUMERIC,
                address=address,
                data=8,
                scope=CommandScope.ENGINE,
            )
            self.flicker_off_btn.when_pressed = cmd.as_action()
        else:
            self.flicker_off_btn = None



from __future__ import annotations

from .sequence_constants import SequenceCommandEnum
from .sequence_req import SequenceReq, T
from ..constants import CommandScope
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum, tmcc2_speed_to_rpm


class SpeedReq(SequenceReq):
    def __init__(
        self,
        address: int,
        speed: int | str | T = None,
        scope: CommandScope = CommandScope.ENGINE,
        is_tmcc: bool = False,
    ) -> None:
        super().__init__(SequenceCommandEnum.ABSOLUTE_SPEED_SEQ, address, scope)
        t, s, sp, e = self.decode_rr_speed(speed, is_tmcc)
        self.add(t, address, scope=scope)
        self.add(s, address, scope=scope, delay=3)
        if is_tmcc is False:
            rpm = tmcc2_speed_to_rpm(sp)
            self.add(TMCC2EngineCommandEnum.DIESEL_RPM, address, data=rpm, scope=scope, delay=4)
        self.add(e, address, scope=scope, delay=6)
